import pytest
from .service_config import ServiceConfig
from aikido_zen.helpers.iplist import IPList


def test_service_config_initialization():
    endpoints = [
        {
            "graphql": False,
            "method": "POST",
            "route": "/v1",
            "rate_limiting": {
                "enabled": True,
                "max_requests": 10,
                "window_size_in_ms": 1000,
            },
            "force_protection_off": False,
        },
        {
            "graphql": True,
            "method": "POST",
            "route": "/v2",
            "rate_limiting": {
                "enabled": True,
                "max_requests": 10,
                "window_size_in_ms": 1000,
            },
            "force_protection_off": False,
        },
        {
            "graphql": False,
            "method": "POST",
            "route": "/v3",
            "rate_limiting": {
                "enabled": True,
                "max_requests": 10,
                "window_size_in_ms": 1000,
            },
            "force_protection_off": False,
        },
        {
            "graphql": False,
            "method": "GET",
            "route": "/admin",
            "rate_limiting": {
                "enabled": False,
                "max_requests": 10,
                "window_size_in_ms": 1000,
            },
            "allowedIPAddresses": ["1.2.3.4", "192.168.2.0/24"],
            "force_protection_off": False,
        },
    ]
    last_updated_at = "2023-10-01"
    service_config = ServiceConfig(
        endpoints,
        last_updated_at,
        ["0", "0", "1", "5"],
        ["127.0.0.1", "123.1.2.0/24", "132.1.0.0/16"],
        True,
    )

    # Check that non-GraphQL endpoints are correctly filtered
    assert len(service_config.endpoints) == 3
    assert service_config.endpoints[0]["route"] == "/v1"
    assert service_config.endpoints[1]["route"] == "/v3"
    assert service_config.endpoints[2]["route"] == "/admin"
    assert service_config.last_updated_at == last_updated_at
    assert isinstance(service_config.bypassed_ips, IPList)
    assert service_config.bypassed_ips.matches("127.0.0.1")
    assert service_config.bypassed_ips.matches("123.1.2.2")
    assert not service_config.bypassed_ips.matches("1.1.1.1")
    assert service_config.blocked_uids == set(["1", "0", "5"])

    v1_endpoint = service_config.get_endpoints(
        {
            "method": "POST",
            "route": "/v1",
            "url": "/v1",
        }
    )[0]
    assert v1_endpoint["route"] == "/v1"
    assert not "allowedIPAddresses" in v1_endpoint

    admin_endpoint = service_config.get_endpoints(
        {
            "method": "GET",
            "route": "/admin",
            "url": "/admin",
        }
    )[0]
    assert admin_endpoint["route"] == "/admin"
    assert isinstance(admin_endpoint["allowedIPAddresses"], IPList)
    assert admin_endpoint["allowedIPAddresses"].matches("192.168.2.1")
    assert admin_endpoint["allowedIPAddresses"].matches("1.2.3.4")
    assert not admin_endpoint["allowedIPAddresses"].matches("192.168.0.1")


# Sample data for testing
sample_endpoints = [
    {"url": "http://example.com/api/v1", "graphql": False, "context": "user"},
    {"url": "http://example.com/api/v2", "graphql": True, "context": "admin"},
    {"url": "http://example.com/api/v3", "graphql": False, "context": "guest"},
]


@pytest.fixture
def service_config():
    return ServiceConfig(
        endpoints=sample_endpoints,
        last_updated_at="2023-10-01T00:00:00Z",
        blocked_uids=["user1", "user2"],
        bypassed_ips=["192.168.1.1", "10.0.0.1"],
        received_any_stats=True,
    )


def test_initialization(service_config):
    assert len(service_config.endpoints) == 2  # Only non-graphql endpoints
    assert service_config.last_updated_at == "2023-10-01T00:00:00Z"
    assert isinstance(service_config.bypassed_ips, IPList)
    assert service_config.blocked_uids == {"user1", "user2"}


def test_ip_blocking():
    config = ServiceConfig(
        endpoints=sample_endpoints,
        last_updated_at="2023-10-01T00:00:00Z",
        blocked_uids=["user1", "user2"],
        bypassed_ips=["192.168.1.1", "10.0.0.0/16", "::1/128"],
        received_any_stats=True,
    )
    config.set_blocked_ips(
        [
            {
                "source": "geoip",
                "description": "description",
                "ips": [
                    "1.2.3.4",
                    "192.168.2.1/24",
                    "fd00:1234:5678:9abc::1",
                    "fd00:3234:5678:9abc::1/64",
                    "5.6.7.8/32",
                ],
            }
        ]
    )

    assert config.is_blocked_ip("1.2.3.4") is "description"
    assert config.is_blocked_ip("2.3.4.5") is False
    assert config.is_blocked_ip("192.168.2.2") is "description"
    assert config.is_blocked_ip("fd00:1234:5678:9abc::1") is "description"
    assert config.is_blocked_ip("fd00:1234:5678:9abc::2") is False
    assert config.is_blocked_ip("fd00:3234:5678:9abc::1") is "description"
    assert config.is_blocked_ip("fd00:3234:5678:9abc::2") is "description"
    assert config.is_blocked_ip("5.6.7.8") is "description"
    assert config.is_blocked_ip("1.2") is False

    assert config.is_bypassed_ip("192.168.1.1")
    assert config.is_bypassed_ip("10.0.0.1")
    assert config.is_bypassed_ip("10.0.1.2")
    assert config.is_bypassed_ip("10.0.254.254")
    assert config.is_bypassed_ip("::1")
    assert not config.is_bypassed_ip("::2")
    assert not config.is_bypassed_ip("1.1.1.1")
    assert not config.is_bypassed_ip("10.1.0.0")


def test_service_config_with_empty_allowlist():
    endpoints = [
        {
            "graphql": False,
            "method": "GET",
            "route": "/admin",
            "rate_limiting": {
                "enabled": False,
                "max_requests": 10,
                "window_size_in_ms": 1000,
            },
            "allowedIPAddresses": [],
            "force_protection_off": False,
        },
    ]
    last_updated_at = "2023-10-01"
    service_config = ServiceConfig(
        endpoints,
        last_updated_at,
        ["0", "0", "1", "5"],
        ["127.0.0.1", "123.1.2.0/24", "132.1.0.0/16"],
        True,
    )

    # Check that non-GraphQL endpoints are correctly filtered
    assert len(service_config.endpoints) == 1
    assert service_config.endpoints[0]["route"] == "/admin"
    assert service_config.last_updated_at == last_updated_at
    assert isinstance(service_config.bypassed_ips, IPList)
    assert service_config.bypassed_ips.matches("127.0.0.1")
    assert service_config.bypassed_ips.matches("123.1.2.2")
    assert not service_config.bypassed_ips.matches("1.1.1.1")
    assert service_config.blocked_uids == set(["1", "0", "5"])

    admin_endpoint = service_config.get_endpoints(
        {
            "method": "GET",
            "route": "/admin",
            "url": "/admin",
        }
    )[0]
    assert admin_endpoint["route"] == "/admin"
    assert isinstance(admin_endpoint["allowedIPAddresses"], list)
    assert len(admin_endpoint["allowedIPAddresses"]) == 0



import pytest
from .percentiles import percentiles


def generate_array_simple(amount):
    return [i + 1 for i in range(amount)]


def shuffle_array(arr):
    import random

    random.shuffle(arr)
    return arr


stubs_simple = [
    {"percentile": 0, "list": shuffle_array(generate_array_simple(100)), "result": 1},
    {"percentile": 25, "list": shuffle_array(generate_array_simple(100)), "result": 25},
    {"percentile": 50, "list": shuffle_array(generate_array_simple(100)), "result": 50},
    {"percentile": 75, "list": shuffle_array(generate_array_simple(100)), "result": 75},
    {
        "percentile": 100,
        "list": shuffle_array(generate_array_simple(100)),
        "result": 100,
    },
    {
        "percentile": 75,
        "list": shuffle_array(generate_array_simple(100) + generate_array_simple(30)),
        "result": 68,
    },
]


def test_percentile_of_simple_values():
    for stub in stubs_simple:
        assert percentiles([stub["percentile"]], stub["list"]) == [stub["result"]]


def test_percentile_with_negative_values():
    assert percentiles([50], shuffle_array([-1, -2, -3, -4, -5])) == [-3]
    assert percentiles([50], shuffle_array([7, 6, -1, -2, -3, -4, -5])) == [-2]


def test_array_of_percentiles():
    assert percentiles(
        [0, 25, 50, 75, 100], shuffle_array(generate_array_simple(100))
    ) == [1, 25, 50, 75, 100]


def test_throw_error_if_less_than_zero():
    with pytest.raises(ValueError):
        percentiles([-1], [1])


def test_throw_error_if_greater_than_100():
    with pytest.raises(ValueError):
        percentiles([101], [1])


def test_empty_list():
    with pytest.raises(ValueError):
        percentiles([50], [])



"""
Mainly exports detect_shell_injection function
"""

from .contains_shell_syntax import contains_shell_syntax
from .is_safely_encapsulated import is_safely_encapsulated


def detect_shell_injection(command, user_input):
    """
    Detects if there is a shell injection give user input
    and the command
    """
    # Block single ~ character. For example echo ~
    if user_input == "~":
        if len(command) > 1 and "~" in command:
            return True

    if len(user_input) <= 1:
        # We ignore single characters since they don't pose a big threat.
        # They are only able to crash the shell, not execute arbitrary commands.
        return False

    if len(user_input) > len(command):
        # We ignore cases where the user input is longer than the command.
        # Because the user input can't be part of the command.
        return False

    if user_input not in command:
        return False

    if is_safely_encapsulated(command, user_input):
        return False

    return contains_shell_syntax(command, user_input)



import copy


class AIStatistics:
    def __init__(self):
        self.calls = {}

    def ensure_provider_stats(self, provider, model):
        key = get_provider_key(provider, model)

        if key not in self.calls:
            self.calls[key] = {
                "provider": provider,
                "model": model,
                "calls": 0,
                "tokens": {
                    "input": 0,
                    "output": 0,
                    "total": 0,
                },
            }

        return self.calls[key]

    def on_ai_call(self, provider, model, input_tokens, output_tokens):
        if not provider or not model:
            return

        provider_stats = self.ensure_provider_stats(provider, model)
        provider_stats["calls"] += 1
        provider_stats["tokens"]["input"] += input_tokens
        provider_stats["tokens"]["output"] += output_tokens
        provider_stats["tokens"]["total"] += input_tokens + output_tokens

    def get_stats(self):
        return [copy.deepcopy(stats) for stats in self.calls.values()]

    def import_list(self, ai_stats_list):
        for new_entry in ai_stats_list:
            existing_entry = self.ensure_provider_stats(
                new_entry["provider"], new_entry["model"]
            )
            existing_entry["calls"] += new_entry["calls"]
            existing_entry["tokens"]["input"] += new_entry["tokens"]["input"]
            existing_entry["tokens"]["output"] += new_entry["tokens"]["output"]
            existing_entry["tokens"]["total"] += new_entry["tokens"]["total"]

    def clear(self):
        self.calls.clear()

    def empty(self):
        return len(self.calls) == 0


def get_provider_key(provider, model):
    return f"{provider}:{model}"



"""
Only exports find_hostname_in_userinput function
"""

from aikido_zen.helpers.get_port_from_url import get_port_from_url
from aikido_zen.helpers.try_parse_url import try_parse_url


def find_hostname_in_userinput(user_input, hostname, port=None):
    """
    Returns true if the hostname is in userinput
    """
    if len(user_input) <= 1:
        return False

    hostname_url = try_parse_url(f"http://{hostname}")
    if not hostname_url:
        return False

    variants = [user_input, f"http://{user_input}", f"https://{user_input}"]
    for variant in variants:
        user_input_url = try_parse_url(variant)
        if user_input_url and user_input_url.hostname == hostname_url.hostname:
            user_port = get_port_from_url(user_input_url.geturl())

            # We were unable to retrieve the port from the URL, likely because it contains an invalid port.
            # Let's assume we have found the hostname in the user input, even though it doesn't match on port.
            # See: https://github.com/AikidoSec/firewall-python/pull/180.
            if user_port is None:
                return True

            if port is None:
                return True
            if port is not None and user_port == port:
                return True

    return False



from pathlib import Path, PurePath
import pytest
from .path_to_string import path_to_string


def test_path_to_string_with_string():
    assert path_to_string("test.txt") == "test.txt"
    assert path_to_string("/home/user/file.txt") == "/home/user/file.txt"


def test_path_to_string_with_valid_url():
    assert path_to_string("file:///path/to/resource") == "/path/to/resource"
    assert path_to_string("file:///another/path") == "/another/path"


def test_path_to_string_with_bytes():
    assert path_to_string(b"test.txt") == "test.txt"
    assert path_to_string(b"/home/user/file.txt") == "/home/user/file.txt"
    assert path_to_string(b"\xff") is None  # Invalid UTF-8 byte sequence


def test_path_to_string_with_empty_string():
    assert path_to_string("") == ""


def test_path_to_string_with_none():
    assert path_to_string(None) is None


def test_path_to_string_with_other_types():
    assert path_to_string(123) is None  # Integer input
    assert path_to_string([]) is None  # List input
    assert path_to_string({}) is None  # Dictionary input


def test_path_to_string_with_pure_path():
    assert path_to_string(PurePath("./", "/folder", "/test.py")) == "/test.py"
    assert path_to_string(PurePath("./", "/folder", "test2.py")) == "/folder/test2.py"
    assert path_to_string(PurePath(".", ".", ".")) == "."
    assert path_to_string(PurePath()) == "."
    assert path_to_string(PurePath("test1", "test2", "test3")) == "test1/test2/test3"


def test_path_to_string_with_path():
    assert path_to_string(Path("./", "/folder", "/test.py")) == "/test.py"
    assert path_to_string(Path("./", "/folder", "test2.py")) == "/folder/test2.py"
    assert path_to_string(Path(".", ".", ".")) == "."
    assert path_to_string(Path()) == "."
    assert path_to_string(Path("test1", "test2", "test3")) == "test1/test2/test3"



"""Helper functions for packages"""

import importlib.metadata as importlib_metadata

from packaging.version import Version
from aikido_zen.helpers.logging import logger

# If any version is supported, this constant can be used
ANY_VERSION = "0.0.0"


def is_package_compatible(package=None, required_version=ANY_VERSION, packages=None):
    """Checks for compatibility of one or multiple package names (in the case of psycopg, you need to check multiple names)"""
    # Fetch package version :
    if package is not None:
        packages = [package]
    if packages is None:
        return False  # no package names provided, return false.
    try:
        for package in packages:
            # Checks if we already looked up the package :
            if PackagesStore.get_package(package) is not None:
                return PackagesStore.get_package(package)["supported"]

            # Safely get the package version, with an exception for when the package was not found
            try:
                package_version = importlib_metadata.version(package)
            except importlib_metadata.PackageNotFoundError:
                continue

            # Check support and store package for later
            supported = is_version_supported(package_version, required_version)
            PackagesStore.add_package(package, package_version, supported)

            if supported:
                logger.debug(
                    "Instrumentation for %s=%s supported", package, package_version
                )
                return True

        # No match found
        logger.info("Zen does not support current version of %s", "/".join(packages))
        return False
    except Exception as e:
        logger.debug("Exception occurred in is_package_compatible: %s", e)
        return False  # We don't support it, since something unexpected happened in checking compatibility.


def is_version_supported(version, required_version):
    """Checks if the package version is supported"""
    return Version(version) >= Version(required_version)


# packages store, uses python's built in GlobalInterpreterLock (GIL)
packages = dict()


class PackagesStore:
    @staticmethod
    def get_packages():
        global packages
        return packages

    @staticmethod
    def add_package(package, version, supported):
        global packages
        packages[package] = {
            "version": version,
            "supported": bool(supported),
        }

    @staticmethod
    def get_package(package_name):
        global packages
        if package_name in packages:
            return packages[package_name]
        return None



import pytest
from aikido_zen.context import Context
from .context_contains_sql_injection import context_contains_sql_injection


class Context2(Context):
    def __init__(self):
        self.cookies = {}
        self.headers = {}
        self.remote_address = "ip"
        self.method = "POST"
        self.url = "url"
        self.body = {}
        self.query = {
            "domain": "www.example`whoami`.com",
        }
        self.source = "express"
        self.route = "/"
        self.parsed_userinput = {}


@pytest.mark.parametrize(
    "invalid_input",
    [
        None,
        123,  # Integer
        45.67,  # Float
        [],  # Empty list
        [1, 2, 3],  # List of integers
        {},  # Empty dictionary
        {"key": "value"},  # Dictionary
        set(),  # Empty set
        {1, 2, 3},  # Set of integers
        object(),  # Instance of a generic object
        lambda x: x,  # Lambda function
        (1, 2),  # Tuple
        b"bytes",  # Bytes
    ],
)
def test_doesnt_crash_with_invalid_sql(invalid_input):
    context = Context2()
    result = context_contains_sql_injection(
        sql=invalid_input,
        operation="mysqlclient.query",
        context=context,
        dialect="mysql",
    )
    assert result == {}



"""
Logging helper, this will create and set logging
"""

import logging
import sys
import os


APPLICATION_NAME = "Zen"

# Set log level
aikido_debug_env = os.getenv("AIKIDO_DEBUG")
LOG_LEVEL = logging.INFO
if aikido_debug_env is not None:
    debug = aikido_debug_env.lower() in ["true", "1"]
    LOG_LEVEL = logging.DEBUG if debug else logging.INFO

# Create a logger
logger = logging.getLogger(APPLICATION_NAME)

# Configure format
fmt = logging.Formatter(
    "%(name)s: %(asctime)s | %(levelname)s | %(processName)s > %(message)s",
    datefmt="%H:%M:%S",
)
# Get stdout handler
stdout = logging.StreamHandler(stream=sys.stdout)
stdout.setLevel(LOG_LEVEL)
stdout.setFormatter(fmt)
logger.addHandler(stdout)

# Configure logger
logger.setLevel(LOG_LEVEL)



"""Exports function extract_data_from_request"""

from aikido_zen.context import get_current_context


async def extract_data_from_request(request):
    """Extracts json, form_data or body from Starlette request"""
    context = get_current_context()
    if not context:
        return

    # Parse data
    try:
        context.set_body(await request.json())
    except ValueError:
        # Throws error if the body is not json
        pass
    if not context.body:
        form_data = await request.form()
        if form_data:
            # Convert to dict object :
            context.set_body({key: value for key, value in form_data.items()})
    if not context.body:
        context.set_body(await request.body())

    context.set_as_current_context()



"""Exports extract_strings_from_context function"""

from aikido_zen.context import UINPUT_SOURCES as SOURCES
from .extract_strings_from_user_input import extract_strings_from_user_input_cached


def extract_strings_from_context(context):
    """
    Iterates over all sources and generates array :
    [(userinput, path, source), ...]
    """
    results = []
    for source in SOURCES:
        if hasattr(context, source):
            user_inputs = extract_strings_from_user_input_cached(
                getattr(context, source), source
            )
            for user_input, path in user_inputs.items():
                if user_input is not None and path is not None:
                    results.append((user_input, path, source))
    return results



import pytest
from unittest.mock import MagicMock, patch
from . import run_vulnerability_scan
from aikido_zen.context import current_context, Context
from aikido_zen.errors import AikidoSQLInjection
from aikido_zen.thread.thread_cache import get_cache
from aikido_zen.helpers.iplist import IPList


@pytest.fixture(autouse=True)
def run_around_tests():
    get_cache().reset()
    yield
    # Make sure to reset context and cache after every test so it does not
    # interfere with other tests
    current_context.set(None)
    get_cache().reset()


@pytest.fixture
def mock_comms():
    """Fixture to mock the get_comms function."""
    with patch("aikido_zen.background_process.comms.get_comms") as mock:
        yield mock


@pytest.fixture
def get_context():
    wsgi_request = {
        "REQUEST_METHOD": "GET",
        "HTTP_HEADER_1": "header 1 value",
        "HTTP_HEADER_2": "Header 2 value",
        "RANDOM_VALUE": "Random value",
        "HTTP_COOKIE": "sessionId=abc123xyz456;",
        "wsgi.url_scheme": "http",
        "HTTP_HOST": "localhost:8080",
        "PATH_INFO": "/hello",
        "QUERY_STRING": "user=JohnDoe&age=30&age=35",
        "CONTENT_TYPE": "application/json",
        "REMOTE_ADDR": "198.51.100.23",
    }
    context = Context(
        req=wsgi_request,
        body={"test_input_sql": "doggoss2', TRUE"},
        source="flask",
    )
    context.route_params = {"test_input2": "cattss2', TRUE"}
    return context


def test_run_vulnerability_scan_no_context(caplog):
    current_context.set(None)
    get_cache().cache = 1
    run_vulnerability_scan(kind="test", op="test", args=tuple())
    assert len(caplog.text) == 0


def test_run_vulnerability_scan_no_context_no_lifecycle(caplog):
    current_context.set(None)
    get_cache().cache = None
    run_vulnerability_scan(kind="test", op="test", args=tuple())
    assert len(caplog.text) == 0


def test_run_vulnerability_scan_context_no_lifecycle(caplog):
    with pytest.raises(Exception):
        current_context.set(1)
        get_cache().cache = None
        run_vulnerability_scan(kind="test", op="test", args=tuple())


def test_lifecycle_cache_ok(caplog, get_context):
    get_context.set_as_current_context()
    run_vulnerability_scan(kind="test", op="test", args=tuple())
    assert "Vulnerability type test currently has no scans implemented" in caplog.text


def test_ssrf(caplog, get_context):
    current_context.set(None)
    run_vulnerability_scan(kind="ssrf", op="test", args=tuple())


def test_lifecycle_cache_bypassed_ip(caplog, get_context):
    get_context.set_as_current_context()
    cache = get_cache()
    cache.config.bypassed_ips = IPList()
    cache.config.bypassed_ips.add("198.51.100.23")
    assert cache.is_bypassed_ip("198.51.100.23")
    run_vulnerability_scan(kind="test", op="test", args=tuple())
    assert len(caplog.text) == 0


def test_sql_injection(caplog, get_context, monkeypatch):
    get_context.set_as_current_context()
    monkeypatch.setenv("AIKIDO_BLOCK", "1")

    assert get_cache().stats.get_record()["requests"]["attacksDetected"]["total"] == 0
    with pytest.raises(AikidoSQLInjection):
        run_vulnerability_scan(
            kind="sql_injection",
            op="test_op",
            args=("INSERT * INTO VALUES ('doggoss2', TRUE);", "mysql"),
        )
    assert get_cache().stats.get_record()["requests"]["attacksDetected"]["total"] == 1
    assert get_cache().stats.get_record()["requests"]["attacksDetected"]["blocked"] == 1


def test_sql_injection_but_blocking_off(caplog, get_context, monkeypatch):
    get_context.set_as_current_context()
    monkeypatch.setenv("AIKIDO_BLOCK", "0")

    assert get_cache().stats.get_record()["requests"]["attacksDetected"]["total"] == 0
    run_vulnerability_scan(
        kind="sql_injection",
        op="test_op",
        args=("INSERT * INTO VALUES ('doggoss2', TRUE);", "mysql"),
    )
    assert get_cache().stats.get_record()["requests"]["attacksDetected"]["total"] == 1
    assert get_cache().stats.get_record()["requests"]["attacksDetected"]["blocked"] == 0


def test_sql_injection_with_route_params(caplog, get_context, monkeypatch):
    get_context.set_as_current_context()
    monkeypatch.setenv("AIKIDO_BLOCK", "1")
    with pytest.raises(AikidoSQLInjection):
        run_vulnerability_scan(
            kind="sql_injection",
            op="test_op",
            args=("INSERT * INTO VALUES ('cattss2', TRUE);", "mysql"),
        )


def test_sql_injection_with_comms(caplog, get_context, monkeypatch):
    get_context.set_as_current_context()
    monkeypatch.setenv("AIKIDO_BLOCK", "1")
    with patch("aikido_zen.background_process.comms.get_comms") as mock_get_comms:
        # Create a mock comms object
        mock_comms = MagicMock()
        mock_get_comms.return_value = mock_comms  # Set the return value of get_comms
        with pytest.raises(AikidoSQLInjection):
            run_vulnerability_scan(
                kind="sql_injection",
                op="test_op",
                args=("INSERT * INTO VALUES ('doggoss2', TRUE);", "mysql"),
            )
        mock_comms.send_data_to_bg_process.assert_called_once()
        call_args = mock_comms.send_data_to_bg_process.call_args[0]
        assert call_args[0] == "ATTACK"
        assert call_args[1][0]["kind"] == "sql_injection"
        assert (
            call_args[1][0]["metadata"]["sql"]
            == "INSERT * INTO VALUES ('doggoss2', TRUE);"
        )
        assert call_args[1][0]["metadata"]["dialect"] == "mysql"


def test_ssrf_vulnerability_scan_adds_hostname(get_context):
    get_context.set_as_current_context()

    dns_results = MagicMock()
    hostname = "example.com"
    port = 80

    run_vulnerability_scan(kind="ssrf", op="test", args=(dns_results, hostname, port))

    # Verify that hostnames.add was called with the correct arguments
    assert get_cache().hostnames.as_array() == [
        {"hits": 1, "hostname": "example.com", "port": 80}
    ]
    run_vulnerability_scan(kind="ssrf", op="test", args=(dns_results, hostname, port))
    assert get_cache().hostnames.as_array() == [
        {"hits": 2, "hostname": "example.com", "port": 80}
    ]


def test_ssrf_vulnerability_scan_no_port(get_context):
    get_context.set_as_current_context()

    dns_results = MagicMock()
    hostname = "example.com"
    port = 0  # Port is zero, should not add to hostnames

    assert get_cache().stats.get_record()["requests"]["attacksDetected"]["total"] == 0
    run_vulnerability_scan(kind="ssrf", op="test", args=(dns_results, hostname, port))
    assert get_cache().stats.get_record()["requests"]["attacksDetected"]["total"] == 0

    assert get_cache().hostnames.as_array() == []


def test_ssrf_vulnerability_scan_bypassed_ip(get_context):
    get_context.set_as_current_context()
    get_cache().config.bypassed_ips = IPList()
    get_cache().config.bypassed_ips.add("198.51.100.23")

    dns_results = MagicMock()
    hostname = "example.com"
    port = 80

    assert get_cache().stats.get_record()["requests"]["attacksDetected"]["total"] == 0
    run_vulnerability_scan(kind="ssrf", op="test", args=(dns_results, hostname, port))
    assert get_cache().stats.get_record()["requests"]["attacksDetected"]["total"] == 0

    # Verify that hostnames.add was not called due to bypassed IP
    assert get_cache().hostnames.as_array() == []



import os
import pytest
from aikido_zen.context import Context, current_context
from aikido_zen.thread.thread_cache import ThreadCache, get_cache
from aikido_zen.errors import AikidoSSRF
from aikido_zen.background_process.comms import reset_comms
import aikido_zen.sinks.socket
import aikido_zen.sinks.http_client
import requests

SSRF_TEST = "http://ssrf-redirects.testssandbox.com/ssrf-test"
SSRF_TEST_DOMAIN = "http://ssrf-redirects.testssandbox.com/ssrf-test-domain"
SSRF_TEST_TWICE = "http://ssrf-redirects.testssandbox.com/ssrf-test-twice"
SSRF_TEST_DOMAIN_TWICE = "http://ssrf-redirects.testssandbox.com/ssrf-test-domain-twice"

CROSS_DOMAIN_TEST = "http://firewallssrfredirects-env-2.eba-7ifve22q.eu-north-1.elasticbeanstalk.com/ssrf-test"
CROSS_DOMAIN_TEST_DOMAIN_TWICE = "http://firewallssrfredirects-env-2.eba-7ifve22q.eu-north-1.elasticbeanstalk.com/ssrf-test-domain-twice"


@pytest.fixture(autouse=True)
def run_around_tests():
    get_cache().reset()
    yield
    # Make sure to reset context and cache after every test so it does not
    # interfere with other tests
    current_context.set(None)
    get_cache().reset()


def set_context_and_lifecycle(url):
    wsgi_request = {
        "REQUEST_METHOD": "GET",
        "HTTP_HEADER_1": "header 1 value",
        "HTTP_HEADER_2": "Header 2 value",
        "RANDOM_VALUE": "Random value",
        "HTTP_COOKIE": "sessionId=abc123xyz456;",
        "wsgi.url_scheme": "http",
        "HTTP_HOST": "localhost:8080",
        "PATH_INFO": "/hello",
        "QUERY_STRING": "user=JohnDoe&age=30&age=35",
        "CONTENT_TYPE": "application/json",
        "REMOTE_ADDR": "198.51.100.23",
    }
    context = Context(
        req=wsgi_request,
        body={
            "url": url,
        },
        source="flask",
    )
    context.set_as_current_context()


def test_srrf_test(monkeypatch):
    reset_comms()
    set_context_and_lifecycle(SSRF_TEST)
    monkeypatch.setenv("AIKIDO_BLOCK", "1")

    with pytest.raises(AikidoSSRF):
        requests.get(SSRF_TEST)


def test_srrf_test_twice(monkeypatch):
    set_context_and_lifecycle(SSRF_TEST_TWICE)
    monkeypatch.setenv("AIKIDO_BLOCK", "1")
    with pytest.raises(AikidoSSRF):
        requests.get(SSRF_TEST_TWICE)


def test_srrf_test_domain(monkeypatch):
    set_context_and_lifecycle(SSRF_TEST_DOMAIN)
    monkeypatch.setenv("AIKIDO_BLOCK", "1")
    with pytest.raises(AikidoSSRF):
        requests.get(SSRF_TEST_DOMAIN)


def test_srrf_test_domain_twice(monkeypatch):
    set_context_and_lifecycle(SSRF_TEST_DOMAIN_TWICE)
    monkeypatch.setenv("AIKIDO_BLOCK", "1")
    with pytest.raises(AikidoSSRF):
        requests.get(SSRF_TEST_DOMAIN_TWICE)


def test_cross_domain(monkeypatch):
    set_context_and_lifecycle(CROSS_DOMAIN_TEST)
    monkeypatch.setenv("AIKIDO_BLOCK", "1")
    with pytest.raises(AikidoSSRF):
        requests.get(CROSS_DOMAIN_TEST)


def test_cross_domain_test_domain_twice(monkeypatch):
    set_context_and_lifecycle(CROSS_DOMAIN_TEST_DOMAIN_TWICE)
    monkeypatch.setenv("AIKIDO_BLOCK", "1")
    with pytest.raises(AikidoSSRF):
        requests.get(CROSS_DOMAIN_TEST_DOMAIN_TWICE)


def test_no_raises_if_diff_url(monkeypatch):
    set_context_and_lifecycle(CROSS_DOMAIN_TEST_DOMAIN_TWICE)
    monkeypatch.setenv("AIKIDO_BLOCK", "1")
    with pytest.raises(requests.exceptions.ConnectionError):
        requests.get(SSRF_TEST_DOMAIN_TWICE)



import pytest
from unittest.mock import MagicMock, patch
from .on_detected_attack import on_detected_attack
from ...context import Context


@pytest.fixture
def mock_connection_manager():
    connection_manager = MagicMock()
    connection_manager.token = "test_token"
    connection_manager.block = True
    connection_manager.timeout_in_sec = 5
    connection_manager.api.report = MagicMock(return_value={"status": "success"})
    connection_manager.get_manager_info = lambda: {}
    return connection_manager


@pytest.fixture
def mock_context():
    basic_wsgi_req = {
        "REQUEST_METHOD": "GET",
        "HTTP_HEADER_1": "header 1 value",
        "HTTP_HEADER_2": "Header 2 value",
        "RANDOM_VALUE": "Random value",
        "HTTP_COOKIE": "sessionId=abc123xyz456;",
        "wsgi.url_scheme": "http",
        "HTTP_HOST": "localhost:8080",
        "PATH_INFO": "/hello",
        "QUERY_STRING": "user=JohnDoe&age=30&age=35",
        "CONTENT_TYPE": "application/json",
        "REMOTE_ADDR": "198.51.100.23",
        "HTTP_USER_AGENT": "Mozilla/5.0",
    }

    return Context(req=basic_wsgi_req, body=123, source="django")


def test_on_detected_attack_no_token(mock_context):
    connection_manager = MagicMock()
    connection_manager.token = None
    on_detected_attack(connection_manager, {}, mock_context, blocked=False, stack=None)
    connection_manager.api.report.assert_not_called()


def test_on_detected_attack_with_long_payload(mock_connection_manager, mock_context):
    long_payload = "x" * 5000  # Create a payload longer than 4096 characters
    attack = {
        "payload": long_payload,
        "metadata": {"test": "1"},
    }

    on_detected_attack(
        mock_connection_manager, attack, mock_context, blocked=False, stack=None
    )
    assert len(attack["payload"]) == 4096  # Ensure payload is truncated
    mock_connection_manager.api.report.assert_called_once()


def test_on_detected_attack_with_long_metadata(mock_connection_manager, mock_context):
    long_metadata = "x" * 5000  # Create metadata longer than 4096 characters
    attack = {
        "payload": {},
        "metadata": {"test": long_metadata},
    }

    on_detected_attack(
        mock_connection_manager, attack, mock_context, blocked=False, stack=None
    )

    assert (
        attack["metadata"]["test"] == long_metadata[:4096]
    )  # Ensure metadata is truncated
    mock_connection_manager.api.report.assert_called_once()


def test_on_detected_attack_success(mock_connection_manager, mock_context):
    attack = {
        "payload": {"key": "value"},
        "metadata": {},
    }

    on_detected_attack(
        mock_connection_manager, attack, mock_context, blocked=False, stack=None
    )
    assert mock_connection_manager.api.report.call_count == 1


def test_on_detected_attack_exception_handling(
    mock_connection_manager, mock_context, caplog
):
    attack = {
        "payload": {"key": "value"},
        "metadata": {"key": "value"},
    }

    # Simulate an exception during the API call
    mock_connection_manager.api.report.side_effect = Exception("API error")

    on_detected_attack(
        mock_connection_manager, attack, mock_context, blocked=False, stack=None
    )

    assert "Failed to report an attack" in caplog.text


def test_on_detected_attack_with_blocked_and_stack(
    mock_connection_manager, mock_context
):
    attack = {
        "payload": {"key": "value"},
        "metadata": {},
    }
    blocked = True
    stack = "sample stack trace"

    on_detected_attack(
        mock_connection_manager, attack, mock_context, blocked=blocked, stack=stack
    )

    # Check that the attack dictionary has the blocked and stack fields set
    assert attack["blocked"] is True
    assert attack["stack"] == stack
    assert mock_connection_manager.api.report.call_count == 1


def test_on_detected_attack_request_data_and_attack_data(
    mock_connection_manager, mock_context
):
    attack = {
        "payload": {"key": "value"},
        "metadata": {"test": "true"},
    }

    on_detected_attack(
        mock_connection_manager, attack, mock_context, blocked=False, stack=None
    )

    # Extract the call arguments for the report method
    _, event, _ = mock_connection_manager.api.report.call_args[0]

    # Verify the request attribute in the payload
    request_data = event["request"]

    assert request_data["method"] == "GET"
    assert request_data["url"] == "http://localhost:8080/hello"
    assert request_data["ipAddress"] == "198.51.100.23"
    assert request_data["body"] == 123
    assert request_data["headers"] == {
        "CONTENT_TYPE": "application/json",
        "USER_AGENT": "Mozilla/5.0",
        "COOKIE": "sessionId=abc123xyz456;",
        "HEADER_1": "header 1 value",
        "HEADER_2": "Header 2 value",
        "HOST": "localhost:8080",
    }
    assert request_data["source"] == "django"
    assert request_data["route"] == "/hello"
    assert request_data["userAgent"] == "Mozilla/5.0"

    attack_data = event["attack"]
    assert attack_data["blocked"] == False
    assert attack_data["metadata"] == {"test": "true"}
    assert attack_data["payload"] == '{"key": "value"}'
    assert attack_data["stack"] is None
    assert attack_data["user"] is None


def test_on_detected_attack_with_user(mock_connection_manager, mock_context):
    attack = {
        "payload": {"key": "value"},
        "metadata": {},
    }
    # Simulate a user in the context
    mock_context.user = "test_user"

    on_detected_attack(
        mock_connection_manager, attack, mock_context, blocked=False, stack=None
    )

    # Extract the call arguments for the report method
    _, event, _ = mock_connection_manager.api.report.call_args[0]

    # Verify the user is included in the attack data
    assert event["attack"]["user"] == "test_user"



"""Helper function file"""

from urllib.parse import urlparse


def try_parse_url(url):
    """Tries to parse the url using urlparse"""
    try:
        parsed_url = urlparse(url)
        if parsed_url.scheme and parsed_url.netloc:
            return parsed_url
        return None
    except Exception:
        return None



import pytest
from .get_auth_types import get_auth_types


class Context:
    def __init__(self, headers={}, cookies={}):
        self.headers = headers
        self.cookies = cookies


def test_detects_authorization_header():
    context1 = Context(headers={"AUTHORIZATION": "Bearer token"})
    assert get_auth_types(context1) == [{"type": "http", "scheme": "bearer"}]

    context2 = Context(headers={"AUTHORIZATION": "Basic base64"})
    assert get_auth_types(context2) == [{"type": "http", "scheme": "basic"}]

    context3 = Context(headers={"AUTHORIZATION": "custom"})
    assert get_auth_types(context3) == [
        {
            "type": "apiKey",
            "in": "header",
            "name": "Authorization",
        }
    ]


def test_detects_api_keys():
    context1 = Context(headers={"X_API_KEY": "token"})
    assert get_auth_types(context1) == [
        {"type": "apiKey", "in": "header", "name": "x-api-key"}
    ]

    context2 = Context(headers={"API_KEY": "token"})
    assert get_auth_types(context2) == [
        {"type": "apiKey", "in": "header", "name": "api-key"}
    ]


def test_detects_auth_cookies():
    context1 = Context(cookies={"api-key": "token"})
    assert get_auth_types(context1) == [
        {"type": "apiKey", "in": "cookie", "name": "api-key"}
    ]

    context2 = Context(cookies={"session": "test"})
    assert get_auth_types(context2) == [
        {
            "type": "apiKey",
            "in": "cookie",
            "name": "session",
        }
    ]


def test_no_auth():
    assert get_auth_types(Context()) is None
    assert get_auth_types(Context(headers={})) is None
    assert get_auth_types(Context(headers={"AUTHORIZATION": ""})) is None
    assert get_auth_types(Context(headers={"AUTHORIZATION": None})) is None



"""Exports `build_url_from_asgi`"""


def build_url_from_asgi(scope):
    """Builds a URL from the different parts of the ASGI scope"""
    scheme = scope["scheme"]

    server = scope["server"]
    if not server:
        return
    # server[0] : Host, server[1] : Port
    host = f"{server[0]}:{server[1]}"

    root_path = scope.get("root_path", "")
    path = scope.get("path", "")
    uri = path.replace(root_path, "", 1)
    return f"{scheme}://{host}{uri}"



"""
Exports the is_safely_encapsulated function
"""

from aikido_zen.helpers.get_current_and_next_segments import (
    get_current_and_next_segments,
)

escape_chars = ['"', "'"]
dangerous_chars_inside_double_quotes = ["$", "`", "\\", "!"]


def is_safely_encapsulated(command, user_input):
    """Checks if the user input is safely encapsulated inside the command"""
    segments = get_current_and_next_segments(command.split(user_input))

    for segment in segments:
        current_segment = segment[0]
        next_segment = segment[1]

        char_before_user_input = current_segment[-1] if current_segment else None
        char_after_user_input = next_segment[0] if next_segment else None

        is_escape_char = char_before_user_input in escape_chars

        if not is_escape_char:
            return False

        if char_before_user_input != char_after_user_input:
            return False

        if char_before_user_input in user_input:
            return False

        #  There are no dangerous characters inside single quotes
        #  You can use certain characters inside double quotes
        #  https://www.gnu.org/software/bash/manual/html_node/Single-Quotes.html
        #  https://www.gnu.org/software/bash/manual/html_node/Double-Quotes.html
        if char_before_user_input == '"' and any(
            char in user_input for char in dangerous_chars_inside_double_quotes
        ):
            return False

    return True



import pytest
from .get_api_info import get_api_info


class Context:
    def __init__(
        self,
        method,
        path,
        body={},
        xml={},
        query={},
        content_type="application/x-www-form-urlencoded",
    ):
        self.method = method
        self.path = path
        self.body = body
        self.xml = xml
        self.headers = {"CONTENT_TYPE": content_type}
        self.query = query
        self.cookies = {}


def test_get_api_info_with_form_encoded_context(monkeypatch):
    monkeypatch.setenv("AIKIDO_FEATURE_COLLECT_API_SCHEMA", "1")
    context1 = Context(
        "GET",
        "/api/resource1",
        body={
            "data1": {
                "data2": [{"Help": True}, {"Help": True, "location": "Sea"}],
                "identifier": "hsfkjewhfwehgkjwehgkj",
                "active": True,
            },
            "user": {"name": "John Doe", "email": "john.doe@example.com"},
        },
    )
    api_info = get_api_info(context1)
    assert api_info == {
        "body": {
            "schema": {
                "properties": {
                    "data1": {
                        "properties": {
                            "active": {
                                "type": "boolean",
                            },
                            "data2": {
                                "items": {
                                    "properties": {
                                        "Help": {
                                            "type": "boolean",
                                        },
                                        "location": {
                                            "optional": True,
                                            "type": "string",
                                        },
                                    },
                                    "type": "object",
                                },
                                "type": "array",
                            },
                            "identifier": {
                                "type": "string",
                            },
                        },
                        "type": "object",
                    },
                    "user": {
                        "properties": {
                            "email": {
                                "type": "string",
                            },
                            "name": {
                                "type": "string",
                            },
                        },
                        "type": "object",
                    },
                },
                "type": "object",
            },
            "type": "form-urlencoded",
        },
        "auth": None,
        "query": None,
    }


def test_get_api_info_with_json(monkeypatch):
    monkeypatch.setenv("AIKIDO_FEATURE_COLLECT_API_SCHEMA", "1")
    context1 = Context(
        "GET",
        "/api/resource1",
        body={
            "user": {"name": "John Doe", "email": "john.doe@example.com"},
        },
        query={
            "user2": {"nickname": "John Doe", "mail": "john.doe@example.com"},
        },
        content_type="application/json",
    )
    api_info = get_api_info(context1)
    assert api_info == {
        "body": {
            "schema": {
                "properties": {
                    "user": {
                        "properties": {
                            "email": {
                                "type": "string",
                            },
                            "name": {
                                "type": "string",
                            },
                        },
                        "type": "object",
                    },
                },
                "type": "object",
            },
            "type": "json",
        },
        "query": {
            "properties": {
                "user2": {
                    "properties": {
                        "mail": {
                            "type": "string",
                        },
                        "nickname": {
                            "type": "string",
                        },
                    },
                    "type": "object",
                },
            },
            "type": "object",
        },
        "auth": None,
    }


def test_auth_get_api_info(monkeypatch):
    monkeypatch.setenv("AIKIDO_FEATURE_COLLECT_API_SCHEMA", "1")
    context1 = Context(
        "GET",
        "/api/resource1",
        body={
            "user": {"name": "John Doe", "email": "john.doe@example.com"},
        },
        query={
            "user2": {"nickname": "John Doe", "mail": "john.doe@example.com"},
        },
        content_type="application/json",
    )
    context1.headers["AUTHORIZATION"] = "Bearer token"
    api_info = get_api_info(context1)
    assert api_info == {
        "body": {
            "schema": {
                "properties": {
                    "user": {
                        "properties": {
                            "email": {
                                "type": "string",
                            },
                            "name": {
                                "type": "string",
                            },
                        },
                        "type": "object",
                    },
                },
                "type": "object",
            },
            "type": "json",
        },
        "query": {
            "properties": {
                "user2": {
                    "properties": {
                        "mail": {
                            "type": "string",
                        },
                        "nickname": {
                            "type": "string",
                        },
                    },
                    "type": "object",
                },
            },
            "type": "object",
        },
        "auth": [{"scheme": "bearer", "type": "http"}],
    }



"""Main export is process_attack"""


def process_attack(connection_manager, data, queue):
    """
    Adds ATTACK data object to queue
    Expected data object : [injection_results, context, blocked_or_not, stacktrace]
    """
    queue.put(data)



"""
Module with logic to build a route, i.e. find route params
from a simple URL string
"""

import re
import ipaddress
from aikido_zen.helpers.looks_like_a_secret import looks_like_a_secret
from aikido_zen.helpers.try_parse_url_path import try_parse_url_path

UUID_REGEX = re.compile(
    r"(?:[0-9a-f]{8}-[0-9a-f]{4}-[1-8][0-9a-f]{3}-[89ab][0-9a-f]{3}-[0-9a-f]{12}|00000000-0000-0000-0000-000000000000|ffffffff-ffff-ffff-ffff-ffffffffffff)$",
    re.I,
)
NUMBER_REGEX = re.compile(r"^\d+$")
DATE_REGEX = re.compile(r"^\d{4}-\d{2}-\d{2}|\d{2}-\d{2}-\d{4}$")
EMAIL_REGEX = re.compile(
    r"^[a-zA-Z0-9.!#$%&\'*+/=?^_`{|}~-]+@[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?(?:\.[a-zA-Z0-9](?:[a-zA-Z0-9-]{0,61}[a-zA-Z0-9])?)*$"
)
HASH_REGEX = re.compile(
    r"^(?:[a-f0-9]{32}|[a-f0-9]{40}|[a-f0-9]{64}|[a-f0-9]{128})$", re.I
)
HASH_LENGTHS = [32, 40, 64, 128]


def build_route_from_url(url):
    """
    Main helper function which will build the route
    from a URL string as input
    """
    path = try_parse_url_path(url)

    if not path:
        return None

    route = "/".join(
        [replace_url_segment_with_param(segment) for segment in path.split("/")]
    )

    if route == "/":
        return "/"

    if route.endswith("/"):
        return route[:-1]

    return route


def replace_url_segment_with_param(segment):
    """
    ??????????
    """
    if not segment:  # Check if segment is empty
        return segment  # Return the segment as is if it's empty
    char_code = ord(segment[0])
    starts_with_number = 48 <= char_code <= 57  # ASCII codes for '0' to '9'

    if starts_with_number and NUMBER_REGEX.match(segment):
        return ":number"

    if len(segment) == 36 and UUID_REGEX.match(segment):
        return ":uuid"

    if starts_with_number and DATE_REGEX.match(segment):
        return ":date"

    if "@" in segment and EMAIL_REGEX.match(segment):
        return ":email"

    try:
        ipaddress.ip_address(segment)
        return ":ip"
    except ValueError:
        pass

    if len(segment) in HASH_LENGTHS and HASH_REGEX.match(segment):
        return ":hash"

    if looks_like_a_secret(segment):
        return ":secret"

    return segment



import pytest
from aikido_zen.helpers.token import Token
from aikido_zen.background_process.api.http_api import ReportingApiHTTP
from aikido_zen.background_process.service_config import ServiceConfig
from aikido_zen.storage.users import Users
from aikido_zen.storage.hostnames import Hostnames
from aikido_zen.ratelimiting.rate_limiter import RateLimiter
from aikido_zen.storage.statistics import Statistics
from . import CloudConnectionManager


@pytest.fixture
def setup_cloud_connection_manager():
    block = None  # Replace with appropriate mock or object if needed
    api = ReportingApiHTTP(None)  # Mock or create an instance of ReportingApiHTTP
    token = Token("AIK_TOKEN_TEST")  # Mock or create an instance of Token
    serverless = "some_value"  # Valid serverless value
    return CloudConnectionManager(block, api, token, serverless)


def test_cloud_connection_manager_initialization(setup_cloud_connection_manager):
    manager = setup_cloud_connection_manager

    # Check that instance variables are initialized correctly
    assert manager.block is None
    assert isinstance(manager.api, ReportingApiHTTP)
    assert isinstance(manager.token, Token)
    assert len(manager.routes) == 0
    assert isinstance(manager.hostnames, Hostnames)
    assert isinstance(manager.conf, ServiceConfig)
    assert isinstance(manager.rate_limiter, RateLimiter)
    assert isinstance(manager.users, Users)
    assert isinstance(manager.statistics, Statistics)
    assert manager.middleware_installed is False
    assert manager.serverless == "some_value"


def test_cloud_connection_manager_empty_serverless():
    block = None  # Replace with appropriate mock or object if needed
    api = ReportingApiHTTP(None)  # Mock or create an instance of ReportingApiHTTP
    token = Token("Hellow")  # Mock or create an instance of Token
    serverless = ""  # Invalid serverless value

    with pytest.raises(ValueError, match="Serverless cannot be an empty string"):
        CloudConnectionManager(block, api, token, serverless)


# Additional tests can be added here for other edge cases or scenarios



import pytest
from .is_http_auth_scheme import is_http_auth_scheme


def test_is_http_auth_scheme():
    assert is_http_auth_scheme("Bearer") == True
    assert is_http_auth_scheme("BEARER") == True
    assert is_http_auth_scheme("DiGeSt") == True
    assert is_http_auth_scheme("invalid") == False



"""
Helper function file, see function docstring
"""

import os


def get_temp_dir():
    """
    Checks the environment variable "AIKIDO_TMP_DIR"
    """
    aikido_temp_dir_env = os.getenv("AIKIDO_TMP_DIR")
    if aikido_temp_dir_env is not None:
        return aikido_temp_dir_env
    return "/tmp"



import pytest
from .extract_wsgi_headers import extract_wsgi_headers


def test_extract_wsgi_headers_single_header():
    request = {"REQUEST_METHOD": "GET", "HTTP_USER_AGENT": "Mozilla/5.0"}
    expected = {"USER_AGENT": "Mozilla/5.0"}
    assert extract_wsgi_headers(request) == expected


def test_extract_wsgi_headers_multiple_headers():
    request = {
        "REQUEST_METHOD": "POST",
        "HTTP_HOST": "example.com",
        "HTTP_ACCEPT": "text/html",
        "HTTP_CONTENT_TYPE": "application/json",
    }
    expected = {
        "HOST": "example.com",
        "ACCEPT": "text/html",
        "CONTENT_TYPE": "application/json",
    }
    assert extract_wsgi_headers(request) == expected


def test_extract_wsgi_headers_no_http_headers():
    request = {"REQUEST_METHOD": "GET", "REMOTE_ADDR": "192.168.1.1"}
    expected = {}
    assert extract_wsgi_headers(request) == expected


def test_extract_wsgi_headers_empty_request():
    request = {}
    expected = {}
    assert extract_wsgi_headers(request) == expected


def test_extract_wsgi_headers_mixed_headers():
    request = {
        "REQUEST_METHOD": "GET",
        "HTTP_USER_AGENT": "Mozilla/5.0",
        "HTTP_ACCEPT_LANGUAGE": "en-US,en;q=0.5",
        "OTHER_HEADER": "value",
    }
    expected = {"USER_AGENT": "Mozilla/5.0", "ACCEPT_LANGUAGE": "en-US,en;q=0.5"}
    assert extract_wsgi_headers(request) == expected



"""
Mainly exports `parse_as_file_url`
"""

from urllib.parse import urlparse, urlunparse
from pathlib import Path


def parse_as_file_url(path):
    """Convert a file path as a URL to a file path."""
    if path.startswith("file:"):
        parsed_url = urlparse(path)
        file_path = Path(parsed_url.path)
    else:
        if not path.startswith("/"):
            path = f"/{path}"
        file_path = Path(path)
        file_url = urlunparse(("file", "", str(file_path), "", "", ""))
        parsed_url = urlparse(file_url)

    normalized_path = Path(parsed_url.path).resolve()

    return str(normalized_path)



import pytest
from .get_port_from_url import get_port_from_url
from urllib.parse import urlparse


def test_get_port_from_url():
    assert get_port_from_url("http://localhost:4000") == 4000
    assert get_port_from_url("http://localhost") == 80
    assert get_port_from_url("https://test.com:8080/test?abc=123") == 8080
    assert get_port_from_url("https://localhost") == 443
    assert get_port_from_url("ftp://localhost") is None
    assert get_port_from_url("http://localhost:1337\\u0000asd.php") is None
    assert get_port_from_url("http://localhost:123123/asd.php") is None


def test_get_port_from_parsed_url():
    assert get_port_from_url(urlparse("http://localhost:4000"), True) == 4000
    assert get_port_from_url(urlparse("http://localhost"), True) == 80
    assert (
        get_port_from_url(urlparse("https://test.com:8080/test?abc=123"), True) == 8080
    )
    assert get_port_from_url(urlparse("https://localhost"), True) == 443
    assert get_port_from_url(urlparse("ftp://localhost"), True) is None
    assert (
        get_port_from_url(urlparse("http://localhost:1337\\u0000asd.php"), True) is None
    )
    assert get_port_from_url(urlparse("http://localhost:123123/asd.php"), True) is None



"""
Sink module for `pymongo`
"""

from aikido_zen.helpers.get_argument import get_argument
import aikido_zen.vulnerabilities as vulns
from . import patch_function, on_import, before
from ..helpers.register_call import register_call


@before
def _func_filter_first(func, instance, args, kwargs):
    """Collection.func(filter, ...)"""
    nosql_filter = get_argument(args, kwargs, 0, "filter")
    if not nosql_filter:
        return

    operation = f"pymongo.collection.Collection.{func.__name__}"
    register_call(operation, "nosql_op")

    vulns.run_vulnerability_scan(
        kind="nosql_injection",
        op=operation,
        args=(nosql_filter,),
    )


@before
def _func_filter_second(func, instance, args, kwargs):
    """Collection.func(..., filter, ...)"""
    nosql_filter = get_argument(args, kwargs, 1, "filter")
    if not nosql_filter:
        return

    operation = f"pymongo.collection.Collection.{func.__name__}"
    register_call(operation, "nosql_op")

    vulns.run_vulnerability_scan(
        kind="nosql_injection",
        op=operation,
        args=(nosql_filter,),
    )


@before
def _func_pipeline(func, instance, args, kwargs):
    """Collection.func(pipeline, ...)"""
    nosql_pipeline = get_argument(args, kwargs, 0, "pipeline")
    if not nosql_pipeline:
        return

    operation = f"pymongo.collection.Collection.{func.__name__}"
    register_call(operation, "nosql_op")

    vulns.run_vulnerability_scan(
        kind="nosql_injection",
        op=operation,
        args=(nosql_pipeline,),
    )


@before
def _bulk_write(func, instance, args, kwargs):
    requests = get_argument(args, kwargs, 0, "requests")

    operation = "pymongo.collection.Collection.bulk_write"
    register_call(operation, "nosql_op")

    # Filter requests that contain "_filter"
    requests_with_filter = [req for req in requests if hasattr(req, "_filter")]
    for request in requests_with_filter:
        vulns.run_vulnerability_scan(
            kind="nosql_injection",
            op=operation,
            args=(request._filter,),
        )


@on_import("pymongo.collection", "pymongo", version_requirement="3.10.0")
def patch(m):
    """
    patching pymongo.collection
    - patches Collection.*(filter, ...)
    - patches Collection.*(..., filter, ...)
    - patches Collection.*(pipeline, ...)
    - patches Collection.bulk_write
    src: https://github.com/mongodb/mongo-python-driver/blob/98658cfd1fea42680a178373333bf27f41153759/pymongo/synchronous/collection.py#L136
    """
    # func(filter, ...)
    patch_function(m, "Collection.replace_one", _func_filter_first)
    patch_function(m, "Collection.update_one", _func_filter_first)
    patch_function(m, "Collection.update_many", _func_filter_first)
    patch_function(m, "Collection.delete_one", _func_filter_first)
    patch_function(m, "Collection.delete_many", _func_filter_first)
    patch_function(m, "Collection.count_documents", _func_filter_first)
    patch_function(m, "Collection.find_one_and_delete", _func_filter_first)
    patch_function(m, "Collection.find_one_and_replace", _func_filter_first)
    patch_function(m, "Collection.find_one_and_update", _func_filter_first)
    patch_function(m, "Collection.find", _func_filter_first)
    patch_function(m, "Collection.find_raw_batches", _func_filter_first)
    # find_one not present in list since find_one calls find function.

    # func(..., filter, ...)
    patch_function(m, "Collection.distinct", _func_filter_second)

    # func(pipeline, ...)
    patch_function(m, "Collection.watch", _func_pipeline)
    patch_function(m, "Collection.aggregate", _func_pipeline)
    patch_function(m, "Collection.aggregate_raw_batches", _func_pipeline)

    # bulk_write
    patch_function(m, "Collection.bulk_write", _bulk_write)



"""Wraps starlette.applications for initial request_handler"""

from aikido_zen.context import Context
from ..functions.request_handler import request_handler
from ...helpers.get_argument import get_argument
from ...sinks import on_import, patch_function, before


@before
def _call(func, instance, args, kwargs):
    scope = get_argument(args, kwargs, 0, "scope")
    if not hasattr(scope, "get") or scope.get("type") != "http":
        return

    new_context = Context(req=scope, source="starlette")
    new_context.set_as_current_context()
    request_handler(stage="init")


@on_import("starlette.applications", "starlette")
def patch(m):
    """
    patching module starlette.applications
    - patches: Starlette.__call__
    """
    patch_function(m, "Starlette.__call__", _call)



"""
Exports class Routes
"""

from aikido_zen.helpers.logging import logger
from aikido_zen.api_discovery.update_route_info import update_route_info
from aikido_zen.api_discovery.get_api_info import get_api_info
from .route_to_key import route_to_key


class Routes:
    """
    Stores all routes
    """

    def __init__(self, max_size=1000):
        self.max_size = max_size
        self.routes = {}

    def initialize_route(self, route_metadata):
        """
        Initializes a route for the first time. `hits_delta_since_sync` counts delta between syncs.
        """
        self.manage_routes_size()
        key = route_to_key(route_metadata)
        self.routes[key] = {
            "method": route_metadata.get("method"),
            "path": route_metadata.get("route"),
            "hits": 0,
            "hits_delta_since_sync": 0,
            "apispec": {},
        }

    def increment_route(self, route_metadata):
        """
        Adds a hit to the route (if it exists) specified in route_metadata.
        route_metadata object includes route, url and method
        """
        key = route_to_key(route_metadata)
        if not self.routes.get(key):
            self.initialize_route(route_metadata)
        # Add hits to
        route = self.routes.get(key)
        route["hits"] += 1
        route["hits_delta_since_sync"] += 1

    def update_route_with_apispec(self, route_metadata, apispec):
        """
        Updates apispec of a given route (or creates it).
        route_metadata object includes route, url and method
        """
        key = route_to_key(route_metadata)
        if not self.routes.get(key):
            return
        update_route_info(apispec, self.routes[key])

    def get(self, route_metadata):
        """Gets you the route entry if it exists using route metadata"""
        key = route_to_key(route_metadata)
        return self.routes.get(key)

    def get_routes_with_hits(self):
        """Gets you all routes with a positive hits delta"""
        result = dict()
        for key, route in self.routes.items():
            if route["hits_delta_since_sync"] <= 0:
                continue  # do not add routes without a hit delta
            result[key] = route
        return result

    def clear(self):
        """Deletes all routes"""
        self.routes = {}

    def manage_routes_size(self):
        """
        Evicts LRU routes if the size is too large
        """
        if len(self.routes) >= self.max_size:
            least_used = [None, float("inf")]
            for key, route in self.routes.items():
                if route.get("hits") < least_used[1]:
                    least_used = [key, route.get("hits")]
            if least_used[0]:
                del self.routes[least_used[0]]

    def __iter__(self):
        return iter(self.routes.values())

    def __len__(self):
        return len(self.routes)



import multiprocessing
import threading

from aikido_zen.context import get_current_context
from aikido_zen.thread.process_worker import aikido_process_worker_thread


def load_worker():
    """
    Loads in a new process worker if one does not already exist for the current process
    """
    if get_current_context() is None:
        return  # don't start a worker if it's not related to a request.

    # The name is aikido-process-worker- + the current PID
    thread_name = "aikido-process-worker-" + str(multiprocessing.current_process().pid)
    if any([thread.name == thread_name for thread in threading.enumerate()]):
        return  # The thread already exists, returning.

    # Create a new daemon thread tht will handle communication to and from background agent
    thread = threading.Thread(target=aikido_process_worker_thread, name=thread_name)
    thread.daemon = True
    thread.start()



"""Helper function file"""

from urllib.parse import urlparse
from pathlib import PurePath


def path_to_string(path):
    """Converts an obj that represents a path into a string"""
    if isinstance(path, str):
        try:
            parsed_url = urlparse(path)
            if parsed_url and parsed_url.scheme == "file":
                return parsed_url.path
        except Exception:
            return None
        return path

    if isinstance(path, bytes):
        try:
            return path.decode("utf-8")
        except UnicodeDecodeError:
            return None
    if isinstance(path, PurePath):
        # Stringify PurePath. This can still allow path traversal but in extremely
        # limited cases so it's safe to just stringify for now.
        return str(path)
    return None



"""
Includes try_parse_url which will try and parse
a url using urlparse.
Includes try_parse_url_path
"""

from urllib.parse import urlparse


def try_parse_url(url):
    """try to parse Url with urlparse"""
    try:
        return urlparse(url)
    except ValueError:
        return None


def try_parse_url_path(url):
    """Try and parse url path"""
    parsed = try_parse_url(f"http://localhost{url}" if url.startswith("/") else url)

    if not parsed or not parsed.scheme:
        return None
    if parsed.path == "":
        return "/"
    return parsed.path



import os
import hashlib
import pytest
from .hash_aikido_token import hash_aikido_token


def test_hash_aikido_token_with_valid_token(monkeypatch):
    """Test the hash_aikido_token function with a valid AIKIDO_TOKEN."""
    # Set the environment variable
    monkeypatch.setenv("AIKIDO_TOKEN", "my_secret_token")

    # Expected hash for "my_secret_token"
    expected_hash = "ddd26ab8c5140fb0dc5bd8cdccd6a0102d09c9bdf2466a5cd718373fd42a17b1"

    # Call the function and assert the result
    assert hash_aikido_token() == expected_hash


def test_hash_aikido_token_with_default(monkeypatch):
    """Test the hash_aikido_token function when AIKIDO_TOKEN is not set."""
    # Ensure the environment variable is not set
    monkeypatch.delenv("AIKIDO_TOKEN", raising=False)

    # Expected hash for the default value
    expected_hash = hashlib.sha256("default".encode("utf-8")).hexdigest()

    # Call the function and assert the result
    assert hash_aikido_token() == expected_hash



"""Exports handle_http_response function"""

from aikido_zen.context import get_current_context
from aikido_zen.helpers.is_redirect_status_code import is_redirect_status_code
from aikido_zen.helpers.try_parse_url import try_parse_url
from .find_hostname_in_context import find_hostname_in_context
from .get_redirect_origin import get_redirect_origin


def normalize_headers(headers):
    """Normalize headers to a consistent format."""
    return {k.lower(): v for k, v in headers.items()}


def handle_http_response(http_response, source):
    """Handles a response object from http"""
    context = get_current_context()
    if not context or not source:
        return
    status_code = http_response.code
    if not status_code or not is_redirect_status_code(status_code):
        #  The status code needs to be a redirect status code
        # Otherwise there is no real danger for SSRF attacks
        return

    raw_headers = getattr(http_response, "headers", {})
    headers = normalize_headers(dict(raw_headers))
    location = headers.get("location", None)
    if not isinstance(location, str):
        return

    parsed_location = try_parse_url(location)
    if not parsed_location:
        return

    add_redirect_to_context(source, parsed_location, context)


def add_redirect_to_context(source, destination, context):
    """
    Adds redirects with user provided hostname / url to the context to prevent
    SSRF attacks with redirects.
    """
    redirect_origin = None

    # Check if the source hostname is in the context - is true if its the first redirect
    # in the chain and the user input is the source
    found = find_hostname_in_context(source.hostname, context, source.port)

    # If the source hostname is not in the context, check if it's a redirect in
    # an already existing chain
    if not found and context.outgoing_req_redirects:
        # Get initial source of the redirect chain (first redirect), if url
        # is part of a redirect chain
        redirect_origin = get_redirect_origin(
            redirects=context.outgoing_req_redirects, url=source
        )

    # Get existing redirects or create a new array
    outgoing_redirects = context.outgoing_req_redirects
    if not outgoing_redirects:
        outgoing_redirects = []

    # If it's 1. a initial redirect with user provided url or 2. a redirect in an existing chain,
    # add it to the context
    if found or redirect_origin:
        outgoing_redirects.append({"source": source, "destination": destination})

        # Update context:
        context.outgoing_req_redirects = outgoing_redirects
        context.set_as_current_context()



"""
Mainly exports `contains_unsafe_path_parts`
"""

dangerous_path_parts = ["../", "..\\"]


def contains_unsafe_path_parts(file_path):
    """Check if the file path contains any dangerous path parts."""
    for dangerous_part in dangerous_path_parts:
        if dangerous_part in file_path:
            return True
    return False



import pytest
from unittest.mock import MagicMock
from .packages import is_package_compatible, ANY_VERSION


@pytest.fixture
def mock_get_comms(mocker):
    """Fixture to mock the get_comms function."""

    class FakeComms:
        def __init__(self):
            self.action = None
            self.obj = None
            self.receive = False

        def send_data_to_bg_process(self, action, obj, receive):
            self.action = action
            self.obj = obj
            self.receive = receive
            return {"success": True, "data": True}

    return FakeComms


def test_pkg_compat_check_success(mock_get_comms, mocker):
    """Test successful wrapping of a package."""
    pkg_name = "example_package"
    pkg_version = "1.0.0"
    mock_class = mock_get_comms()
    # Mock the metadata.version function
    mocker.patch("importlib.metadata.version", return_value=pkg_version)
    mocker.patch(
        "aikido_zen.background_process.comms.get_comms", return_value=mock_class
    )

    # Call the function under test
    is_package_compatible(pkg_name, ANY_VERSION)


def test_pkg_compat_check_retry(mock_get_comms, mocker):
    """Test retry logic when sending data fails."""
    pkg_name = "example_package"
    pkg_version = "1.0.0"
    mock_class = mock_get_comms()  # Create an instance of FakeComms

    # Mock the metadata.version function
    mocker.patch("importlib.metadata.version", return_value=pkg_version)
    # Mock the get_comms function to return the mock_class instance
    mocker.patch(
        "aikido_zen.background_process.comms.get_comms", return_value=mock_class
    )

    # Mock the send_data_to_bg_process method to simulate failure
    mock_class.send_data_to_bg_process = MagicMock(
        return_value={
            "success": False,
            "data": False,
        }
    )

    # Call the function under test
    is_package_compatible(pkg_name, ANY_VERSION)


def test_pkg_compat_check_partial_success(mock_get_comms, mocker):
    """Test that the function stops retrying after a successful attempt."""
    pkg_name = "example_package"
    pkg_version = "1.0.0"
    mock_class = mock_get_comms()  # Create an instance of FakeComms

    # Mock the metadata.version function
    mocker.patch("importlib.metadata.version", return_value=pkg_version)
    # Mock the get_comms function to return the mock_class instance
    mocker.patch(
        "aikido_zen.background_process.comms.get_comms", return_value=mock_class
    )

    # Mock the send_data_to_bg_process method to simulate a failure followed by success
    mock_class.send_data_to_bg_process = MagicMock(
        side_effect=[
            {"success": False, "data": False},  # First attempt fails
            {"success": True, "data": True},  # Second attempt succeeds
        ]
    )

    # Call the function under test
    is_package_compatible(pkg_name, ANY_VERSION)



from . import (
    AikidoException,
    AikidoSQLInjection,
    AikidoNoSQLInjection,
    AikidoRateLimiting,
    AikidoShellInjection,
    AikidoPathTraversal,
    AikidoSSRF,
)


def test_aikido_exception_default_message():
    exception = AikidoException()
    assert str(exception) == "(AikidoException(...), 'Zen has blocked unknown')"


def test_aikido_sql_injection():
    exception = AikidoSQLInjection(dialect="MySQL")
    assert (
        str(exception)
        == "(AikidoSQLInjection(...), 'Zen has blocked an SQL injection, dialect: MySQL')"
    )


def test_aikido_nosql_injection():
    exception = AikidoNoSQLInjection()
    assert (
        str(exception)
        == "(AikidoNoSQLInjection(...), 'Zen has blocked a NoSQL injection')"
    )


def test_aikido_rate_limiting():
    exception = AikidoRateLimiting()
    assert str(exception) == "(AikidoRateLimiting(...), 'You are rate limited by Zen.')"


def test_aikido_shell_injection():
    exception = AikidoShellInjection()
    assert (
        str(exception)
        == "(AikidoShellInjection(...), 'Zen has blocked a shell injection')"
    )


def test_aikido_path_traversal():
    exception = AikidoPathTraversal()
    assert (
        str(exception)
        == "(AikidoPathTraversal(...), 'Zen has blocked a path traversal attack')"
    )


def test_aikido_ssrf():
    exception = AikidoSSRF()
    assert (
        str(exception)
        == "(AikidoSSRF(...), 'Zen has blocked a server-side request forgery')"
    )



"""
Helper function file, see funtion definition
"""


def build_path_to_payload(path_to_payload):
    """
    Create a string so people see the path to where
    the injection took place
    """
    if len(path_to_payload) == 0:
        return "."

    result = ""
    for part in path_to_payload:
        if part["type"] == "object":
            result += "." + part["key"]
        elif part["type"] == "array":
            result += f".[{part['index']}]"
        elif part["type"] == "jwt":
            result += "<jwt>"

    return result



import pytest
from aikido_zen import protect
from aikido_zen.background_process import get_comms, reset_comms


def test_protect_with_django(monkeypatch, caplog):
    protect()

    assert "starting" in caplog.text
    reset_comms()
    assert get_comms() == None



"""
Aggregates from the different modules
"""

# Re-export functions :
from aikido_zen.context.users import set_user
from aikido_zen.middleware import should_block_request

# Import logger
from aikido_zen.helpers.logging import logger

# Import background process
from aikido_zen.background_process import start_background_process

# Load environment variables and constants
from aikido_zen.config import PKG_VERSION
from aikido_zen.helpers.aikido_disabled_flag_active import aikido_disabled_flag_active


def protect(mode="daemon"):
    """
    Mode can be set to :
    - daemon : Default, imports sinks/sources and starts background_process
    - daemon_only: Only starts background process and doesn't wrap
    - daemon_disabled : This will import sinks/sources but won't start a background process
    Protect user's application
    """
    if aikido_disabled_flag_active():
        # Do not run any aikido code when the disabled flag is on
        return
    if mode in ("daemon", "daemon_only"):
        start_background_process()
    if mode == "daemon_only":
        # Do not import sinks/sources
        return
    if mode == "daemon_disabled":
        logger.debug("Not starting the background process, daemon disabled.")

    # Import sources
    import aikido_zen.sources.django
    import aikido_zen.sources.flask
    import aikido_zen.sources.quart
    import aikido_zen.sources.starlette
    import aikido_zen.sources.xml_sources.xml
    import aikido_zen.sources.xml_sources.lxml

    # Import DB Sinks
    import aikido_zen.sinks.pymysql
    import aikido_zen.sinks.mysqlclient
    import aikido_zen.sinks.pymongo
    import aikido_zen.sinks.psycopg2
    import aikido_zen.sinks.psycopg
    import aikido_zen.sinks.asyncpg
    import aikido_zen.sinks.clickhouse_driver

    import aikido_zen.sinks.builtins
    import aikido_zen.sinks.os
    import aikido_zen.sinks.shutil
    import aikido_zen.sinks.io
    import aikido_zen.sinks.http_client
    import aikido_zen.sinks.socket

    # Import shell sinks
    import aikido_zen.sinks.os_system
    import aikido_zen.sinks.subprocess

    # Import AI sinks
    import aikido_zen.sinks.openai

    logger.info("Zen by Aikido v%s starting.", PKG_VERSION)



import pytest
from unittest.mock import MagicMock, patch
from .on_start import on_start


@pytest.fixture
def mock_connection_manager():
    connection_manager = MagicMock()
    connection_manager.token = "test_token"
    connection_manager.timeout_in_sec = 5
    connection_manager.api.report = MagicMock(return_value={"success": True})
    connection_manager.get_manager_info = lambda: {}
    connection_manager.update_service_config = MagicMock()
    return connection_manager


def test_on_start_no_token():
    """Test that no API call is made when the connection_manager has no token."""
    connection_manager = MagicMock()
    connection_manager.token = None
    on_start(connection_manager)
    connection_manager.api.report.assert_not_called()


def test_on_start_success(mock_connection_manager, caplog):
    """Test that the API call is made successfully and the service config is updated."""
    on_start(mock_connection_manager)

    # Check that the API report method was called
    mock_connection_manager.api.report.assert_called_once()

    # Check that the service config was updated
    mock_connection_manager.update_service_config.assert_called_once()

    # Check that the info log was called
    assert "Established connection with Aikido Server" in caplog.text


def test_on_start_failure(mock_connection_manager, caplog):
    """Test that an error is logged when the API call fails."""
    mock_connection_manager.api.report.return_value = {
        "success": False,
        "error": "Some error",
    }

    on_start(mock_connection_manager)

    # Check that the API report method was called
    mock_connection_manager.api.report.assert_called_once()

    # Check that the service config was not updated
    mock_connection_manager.update_service_config.assert_not_called()

    # Check that the error log was called
    assert "Failed to communicate with Aikido Server : Some error" in caplog.text



import pytest
from unittest.mock import MagicMock
from .ip_allowed_to_access_route import ip_allowed_to_access_route
from aikido_zen.helpers.iplist import IPList


def gen_route_metadata(
    url="http://localhost:4000/posts/3", method="GET", route="/posts/:id"
):
    return {"url": url, "route": route, "method": method}


def gen_endpoints(allowed_ip_addresses):
    return [
        {
            "route": "/posts/:id",
            "method": "POST",
            "allowedIPAddresses": (
                IPList(allowed_ip_addresses) if len(allowed_ip_addresses) > 0 else None
            ),
            "force_protection_off": False,
        },
    ]


def test_always_allows_request_if_not_production():
    context = gen_route_metadata()
    endpoints = gen_endpoints(["1.2.3.4"])
    assert ip_allowed_to_access_route("::1", context, endpoints) is True


def test_always_allows_request_if_no_match():
    endpoints = gen_endpoints(["1.2.3.4"])
    modified_context = gen_route_metadata(route="/", method="GET")
    assert ip_allowed_to_access_route("1.2.3.4", modified_context, endpoints) is True


def test_always_allows_request_if_allowed_ip_address():
    endpoints = gen_endpoints(["1.2.3.4", "10.0.0.0/24"])
    modified_context = gen_route_metadata()
    assert ip_allowed_to_access_route("1.2.3.4", modified_context, endpoints) is True
    assert ip_allowed_to_access_route("10.0.0.2", modified_context, endpoints) is True


def test_always_allows_request_if_localhost():
    endpoints = gen_endpoints(["1.2.3.4"])
    modified_context = gen_route_metadata()
    assert ip_allowed_to_access_route("::1", modified_context, endpoints) is True


def test_blocks_request_if_no_ip_address():
    endpoints = gen_endpoints(["1.2.3.4"])
    modified_context = gen_route_metadata()
    assert ip_allowed_to_access_route(None, modified_context, endpoints) is False


def test_allows_request_if_configuration_is_broken():
    endpoints = gen_endpoints({})  # Broken configuration
    modified_context = gen_route_metadata(method="POST", route="/posts/:id")
    assert ip_allowed_to_access_route("3.4.5.6", modified_context, endpoints) is True


def test_allows_request_if_allowed_ip_addresses_is_empty():
    endpoints = gen_endpoints([])
    context = gen_route_metadata()
    assert ip_allowed_to_access_route("3.4.5.6", context, endpoints) is True


def test_blocks_request_if_not_allowed_ip_address():
    endpoints = gen_endpoints(["1.2.3.4", "10.0.0.0/24"])
    modified_context = gen_route_metadata()
    assert ip_allowed_to_access_route("3.4.5.6", modified_context, endpoints) is False
    assert ip_allowed_to_access_route("10.1.1.1", modified_context, endpoints) is False


def test_checks_every_matching_endpoint():
    endpoints = [
        {
            "route": "/posts/:id",
            "method": "POST",
            "allowedIPAddresses": IPList(["3.4.5.6"]),
            "force_protection_off": False,
        },
        {
            "route": "/posts/*",
            "method": "POST",
            "allowedIPAddresses": IPList(["1.2.3.4"]),
            "force_protection_off": False,
        },
    ]
    modified_context = gen_route_metadata(method="POST", route="/posts/:id")
    assert ip_allowed_to_access_route("3.4.5.6", modified_context, endpoints) is False


def test_if_allowed_ips_is_empty_or_broken():
    endpoints = [
        {
            "route": "/posts/:id",
            "method": "POST",
            "allowedIPAddresses": [],
            "force_protection_off": False,
        },
        {
            "route": "/posts/*",
            "method": "POST",
            "allowedIPAddresses": {},  # Broken configuration
            "force_protection_off": False,
        },
        {
            "route": "/posts/*",
            "method": "POST",
            "allowedIPAddresses": IPList(["1.2.3.4"]),
            "force_protection_off": False,
        },
    ]
    modified_context_allowed = gen_route_metadata()
    assert (
        ip_allowed_to_access_route("1.2.3.4", modified_context_allowed, endpoints)
        is True
    )

    modified_context_blocked = gen_route_metadata()
    assert (
        ip_allowed_to_access_route("3.4.5.6", modified_context_blocked, endpoints)
        is False
    )



"""
Includes all the wrappers for gunicorn config file
"""

import aikido_zen


# Run our background process as a child of gunicorn (exits safely)
aikido_zen.protect("daemon_only")


def post_fork(prev_func):
    """
    Aikido decorator for gunicorn config
    Function: post_fork(server, worker)
    """

    def aik_post_fork(server, worker):
        aikido_zen.protect("daemon_disabled")
        prev_func(server, worker)

    return aik_post_fork



"""
Exports `detect_path_traversal`
"""

from .contains_unsafe_path_parts import contains_unsafe_path_parts
from .parse_as_file_url import parse_as_file_url
from .unsafe_path_start import starts_with_unsafe_path


def detect_path_traversal(file_path, user_input, check_path_start=True, is_url=False):
    """Detect potential path traversal vulnerabilities."""
    if len(user_input) <= 1:
        # We ignore single characters since they don't pose a big threat.
        return False

    # Check for URL path traversal
    # Reason: new URL("file:///../../test.txt") => /test.txt
    # The normal check for relative path traversal will fail in this case, because transformed path does not contain ../.
    # For absolute path traversal, we dont need to check the transformed path, because it will always start with /.
    # Also /./ is checked by normal absolute path traversal check (if #219 is merged)
    if is_url and contains_unsafe_path_parts(user_input):
        file_path_from_url = parse_as_file_url(user_input)
        if file_path_from_url and file_path_from_url in file_path:
            return True

    if len(user_input) > len(file_path):
        # We ignore cases where the user input is longer than the file path.
        # Because the user input can't be part of the file path.
        return False

    if user_input not in file_path:
        # We ignore cases where the user input is not part of the file path.
        return False

    if contains_unsafe_path_parts(file_path) and contains_unsafe_path_parts(user_input):
        return True

    if check_path_start:
        return starts_with_unsafe_path(
            file_path, user_input
        )  #  Check for absolute path traversal

    return False



"""Helper function file, see function docstring"""

import os

ENV_PREFIX = "AIKIDO_FEATURE_"


def is_feature_enabled(feature):
    """Checks if `feature` is enabled by checking ENV variables"""
    env_variable_name = ENV_PREFIX + feature.upper()
    env_value = os.getenv(env_variable_name)
    if env_value is not None:
        return env_value.lower() in ["true", "1"]
    return False



"""
Sink module for `builtins`, python's built-in function
"""

from pathlib import PurePath
import aikido_zen.vulnerabilities as vulns
from aikido_zen.helpers.get_argument import get_argument
from aikido_zen.helpers.register_call import register_call
from aikido_zen.sinks import patch_function, on_import, before


@before
def _open(func, instance, args, kwargs):
    filename = get_argument(args, kwargs, 0, "filename")
    if not isinstance(filename, (str, bytes, PurePath)):
        return

    op = "builtins.open"
    register_call(op, "fs_op")

    vulns.run_vulnerability_scan(kind="path_traversal", op=op, args=(filename,))


@on_import("builtins")
def patch(m):
    """
    patching module builtins
    - patches builtins.open
    """
    patch_function(m, "open", _open)



import pytest
from unittest.mock import patch
import aikido_zen.sinks.builtins

kind = "path_traversal"


def test_lzma_lzmafile():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import lzma

        with pytest.raises(FileNotFoundError):
            lzma.LZMAFile("sdnklsdnlkvsv.lzma")
        mock_run_vulnerability_scan.assert_called_with(
            kind="path_traversal", op="builtins.open", args=("sdnklsdnlkvsv.lzma",)
        )

        with pytest.raises(FileNotFoundError):
            lzma.LZMAFile(filename="jsgljlskjglkds.lzma")
        mock_run_vulnerability_scan.assert_called_with(
            kind="path_traversal", op="builtins.open", args=("jsgljlskjglkds.lzma",)
        )

        with pytest.raises(FileNotFoundError):
            lzma.LZMAFile(filename="/kjtkl.lzma")
        mock_run_vulnerability_scan.assert_called_with(
            kind="path_traversal", op="builtins.open", args=("/kjtkl.lzma",)
        )


def test_lzma_open():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import lzma

        with pytest.raises(FileNotFoundError):
            lzma.open("sgsglksjks.lzma")
        mock_run_vulnerability_scan.assert_called_with(
            kind="path_traversal", op="builtins.open", args=("sgsglksjks.lzma",)
        )

        with pytest.raises(FileNotFoundError):
            lzma.open(filename="qjrpoiuzu.lzma")
        mock_run_vulnerability_scan.assert_called_with(
            kind="path_traversal", op="builtins.open", args=("qjrpoiuzu.lzma",)
        )



import pytest
from unittest.mock import patch
import aikido_zen.sinks.subprocess

kind = "shell_injection"


def test_subprocess_call():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import subprocess

        op = "subprocess.Popen"

        subprocess.call(["ls", "-la"], shell=True)
        args = ("ls -la",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        subprocess.call(["cfsknflks"], shell=True)
        args = ("cfsknflks",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        subprocess.call("ls -la", shell=True)
        args = ("ls -la",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)


def test_subprocess_call_not_called():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import subprocess

        subprocess.call(["ls", "-la"])

        subprocess.call(["pwd"], shell=False)

        subprocess.call("ls")

        # Make sure there is no call regarding shell injection:
        assert (
            list(filter(is_shell_injection, mock_run_vulnerability_scan.call_args_list))
            == []
        )


def test_subprocess_run():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import subprocess

        op = "subprocess.Popen"

        subprocess.run(["ls", "-la"], shell=True)
        args = ("ls -la",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        subprocess.run(["cfsknflks"], shell=True)
        args = ("cfsknflks",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)


def test_subprocess_run_not_called():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import subprocess

        subprocess.run(["ls", "-la"])

        subprocess.run(["pwd"], shell=False)

        # Make sure there is no call regarding shell injection:
        assert (
            list(filter(is_shell_injection, mock_run_vulnerability_scan.call_args_list))
            == []
        )


def test_subprocess_check_call():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import subprocess

        op = "subprocess.Popen"

        subprocess.check_call(["ls", "-la"], shell=True)
        args = ("ls -la",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        with pytest.raises(subprocess.CalledProcessError):
            subprocess.check_call(["cfsknflks"], shell=True)
        args = ("cfsknflks",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)


def test_subprocess_check_call_not_called():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import subprocess

        subprocess.check_call(["ls", "-la"], shell=False)

        subprocess.check_call(["whoami"])

        # Make sure there is no call regarding shell injection:
        assert (
            list(filter(is_shell_injection, mock_run_vulnerability_scan.call_args_list))
            == []
        )


def test_subprocess_popen():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import subprocess

        op = "subprocess.Popen"

        subprocess.Popen(["ls", "-la"], shell=True)
        args = ("ls -la",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        subprocess.Popen(["cfsknflks"], shell=True)
        args = ("cfsknflks",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        subprocess.Popen("ls -la", shell=True)
        args = ("ls -la",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        process = subprocess.Popen(args="ls -la", shell=True)
        args = ("ls -la",)
        mock_run_vulnerability_scan.assert_called_with(kind=kind, op=op, args=args)

        process.kill()  # Test class functions
        process.pid  # Access a class attribute


def test_subprocess_popen_not_called():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import subprocess

        subprocess.Popen(["ls", "-la"])

        subprocess.Popen(["whoami"], shell=False)

        subprocess.Popen("pwd", shell=False)

        subprocess.Popen(args="pwd", shell=False)

        subprocess.Popen(args="pwd")

        # Make sure there is no call regarding shell injection:
        assert (
            list(filter(is_shell_injection, mock_run_vulnerability_scan.call_args_list))
            == []
        )


def test_subprocess_check_output():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import subprocess

        op = "subprocess.Popen"

        subprocess.check_output(["ls", "-la"], shell=True)
        args = ("ls -la",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        with pytest.raises(subprocess.CalledProcessError):
            subprocess.check_output(["cfsknflks"], shell=True)
        args = ("cfsknflks",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        with pytest.raises(subprocess.CalledProcessError):
            subprocess.check_output(("tuple", "command"), shell=True)
        args = ("tuple command",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        with pytest.raises(subprocess.CalledProcessError):
            subprocess.check_output({"key": "value"}, shell=True)
        args = ("key",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)

        with pytest.raises(subprocess.CalledProcessError):
            subprocess.check_output({"ke": "value", "key2": "value2"}, shell=True)
        args = ("ke key2",)
        mock_run_vulnerability_scan.assert_any_call(kind=kind, op=op, args=args)


def test_subprocess_check_output_2():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import subprocess

        subprocess.check_output(["ls", "-la"], shell=False)

        with pytest.raises(FileNotFoundError):
            subprocess.check_output(["cfsknflks"])

        with pytest.raises(FileNotFoundError):
            subprocess.check_output(("tuple", "command"), shell=False)

        with pytest.raises(FileNotFoundError):
            subprocess.check_output({"key": "value"})

        with pytest.raises(FileNotFoundError):
            subprocess.check_output({"ke": "value", "key2": "value2"}, shell=False)

        # Make sure there is no call regarding shell injection:
        assert (
            list(filter(is_shell_injection, mock_run_vulnerability_scan.call_args_list))
            == []
        )


def test_subprocess_invalid_input():
    with patch(
        "aikido_zen.vulnerabilities.run_vulnerability_scan"
    ) as mock_run_vulnerability_scan:
        import subprocess

        op = "subprocess.call"
        with pytest.raises(TypeError):
            subprocess.call(shell=True)
        mock_run_vulnerability_scan.assert_not_called()

        with pytest.raises(TypeError):
            subprocess.call(123456789123456789, shell=True)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        with pytest.raises(TypeError):
            subprocess.call(None, shell=True)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test float command
        with pytest.raises(TypeError):
            subprocess.call(3.14, shell=True)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called

        # Test boolean command
        with pytest.raises(TypeError):
            subprocess.call(True, shell=True)
        mock_run_vulnerability_scan.assert_not_called()  # Ensure it was not called


def is_shell_injection(args):
    return args[1]["kind"] == "shell_injection"



"""
Aikido can rate limit urls, IPs, Users
"""

from .get_ratelimited_endpoint import get_ratelimited_endpoint


def should_ratelimit_request(route_metadata, remote_address, user, connection_manager):
    """
    Checks if the request should be ratelimited or not
    route_metadata object includes route, url and method
    """
    endpoints = connection_manager.conf.get_endpoints(route_metadata)
    endpoint = get_ratelimited_endpoint(endpoints, route_metadata["route"])
    if not endpoint:
        return {"block": False}

    max_requests = int(endpoint["rateLimiting"]["maxRequests"])
    windows_size_in_ms = int(endpoint["rateLimiting"]["windowSizeInMS"])
    is_bypassed_ip = connection_manager.conf.is_bypassed_ip(remote_address)

    if is_bypassed_ip:
        return {"block": False}
    if user:
        uid = user["id"]
        method = endpoint.get("method")
        route = endpoint.get("route")

        allowed = connection_manager.rate_limiter.is_allowed(
            f"{method}:{route}:user:{uid}",
            windows_size_in_ms,
            max_requests,
        )
        if not allowed:
            return {"block": True, "trigger": "user"}
        # Do not check IP rate limit if user is set
        return {"block": False}

    if remote_address:
        method = endpoint.get("method")
        route = endpoint.get("route")

        allowed = connection_manager.rate_limiter.is_allowed(
            f"{method}:{route}:ip:{remote_address}",
            windows_size_in_ms,
            max_requests,
        )
        if not allowed:
            return {"block": True, "trigger": "ip"}

    return {"block": False}



"""
Helper function file, see function docstring
"""

from urllib.parse import urlparse


def get_subdomains_from_url(url):
    """
    Returns a list with subdomains from url
    """
    if not isinstance(url, str):
        return []
    host = urlparse(url).hostname
    if not host:
        return []
    parts = host.split(".")
    return parts[:-2]



import re
import pytest
from aikido_zen.helpers import escape_string_regexp


def test_escape_string_regexp():
    assert (
        escape_string_regexp("\\ ^ $ * + ? . ( ) | { } [ ]")
        == "\\\\ \\^ \\$ \\* \\+ \\? \\. \\( \\) \\| \\{ \\} \\[ \\]"
    )


def test_escape_hyphen_pcre():
    assert escape_string_regexp("foo - bar") == "foo \\x2d bar"


def test_escape_hyphen_unicode_flag():
    assert re.match(escape_string_regexp("-"), "-u") is not None



"""Exports is_http_auth_scheme function"""

AUTH_SCHEMES = [
    "basic",
    "bearer",
    "digest",
    "dpop",
    "gnap",
    "hoba",
    "mutual",
    "negotiate",
    "privatetoken",
    "scram-sha-1",
    "scram-sha-256",
    "vapid",
]


def is_http_auth_scheme(scheme):
    """
    Checks if a string is a valid HTTP authentication scheme.
    https://www.iana.org/assignments/http-authschemes/http-authschemes.xhtml
    """
    return scheme.lower() in AUTH_SCHEMES



"""Exports is_redirect_to_private_ip"""

from aikido_zen.helpers.get_port_from_url import get_port_from_url
from .get_redirect_origin import get_redirect_origin
from .find_hostname_in_context import find_hostname_in_context


def is_redirect_to_private_ip(hostname, context, port):
    """
    This function is called before an outgoing request is made.
    It's used to prevent requests to private IP addresses after a redirect with
        a user-supplied URL (SSRF).
    It returns True if the following conditions are met:
    - context.outgoing_req_redirects is set: Inside the context of this incoming request,
        there was a redirect
    - The hostname of the URL contains a private IP address
    - The redirect origin, so the user-supplied hostname and port that caused the first redirect,
        is found in the context of the incoming request
    """
    if context.outgoing_req_redirects:
        redirect_origin = get_redirect_origin(
            context.outgoing_req_redirects, hostname, port
        )
        if redirect_origin:
            origin_hostname = getattr(redirect_origin, "hostname")
            origin_port = get_port_from_url(redirect_origin, parsed=True)
            return find_hostname_in_context(origin_hostname, context, origin_port)

    return None



from aikido_zen.context import Context, get_current_context
import aikido_zen.sources.xml_sources.lxml
import lxml.etree as ET

# Sample XML string for testing
XML_STRING = """<?xml version="1.0"?>
<root>
    <child attr="chill" test="test1" smth="2" smthelse="2">
        <name smth="2" test="test2">Test Name</name>
        <value ok="boomer">42</value>
    </child>
</root>
"""


def parse_xml(xml_string):
    """Parse the given XML string and return the root element."""
    try:
        root = ET.fromstring(xml_string)
        return root
    except ET.ParseError as e:
        raise ValueError(f"Failed to parse XML: {e}")


def set_context(body):
    Context(
        req={
            "REQUEST_METHOD": "GET",
            "HTTP_HEADER_1": "header 1 value",
            "HTTP_HEADER_2": "Header 2 value",
            "RANDOM_VALUE": "Random value",
            "HTTP_COOKIE": "sessionId=abc123xyz456;",
            "wsgi.url_scheme": "http",
            "HTTP_HOST": "localhost:8080",
            "PATH_INFO": "/hello",
            "QUERY_STRING": "user=JohnDoe&age=30&age=35",
            "CONTENT_TYPE": "application/json",
            "REMOTE_ADDR": "198.51.100.23",
            "HTTP_USER_AGENT": "Mozilla/5.0",
        },
        body=body,
        source="django",
    ).set_as_current_context()


def test_parse_xml_with_random_context():
    """Test the XML parsing function."""
    set_context("Body falalalala")
    root = parse_xml(XML_STRING)

    # Check XML parser still works
    assert root[0].find("name").text == "Test Name"

    # Check that no xml attribute was set on body
    context = get_current_context()
    assert context.xml == {}


def test_parse_xml_with_set_context():
    """Test the XML parsing function."""
    set_context(body=XML_STRING)
    root = parse_xml(XML_STRING)

    # Check XML parser still works
    assert root[0].find("name").text == "Test Name"

    # Check that no xml attribute was set on body
    context = get_current_context()
    assert context.xml == {
        "attr": {"chill"},
        "smth": {"2"},
        "smthelse": {"2"},
        "test": {"test1"},
    }



from aikido_zen.helpers.extract_data_from_xml_body import (
    extract_data_from_xml_body,
)
from aikido_zen.helpers.get_argument import get_argument
from aikido_zen.helpers.register_call import register_call
from aikido_zen.sinks import on_import, patch_function, after


@after
def _fromstring(func, instance, args, kwargs, return_value):
    text = get_argument(args, kwargs, 0, "text")
    register_call("xml.etree.ElementTree.fromstring", "deserialize_op")

    extract_data_from_xml_body(user_input=text, root_element=return_value)


@after
def _fromstringlist(func, instance, args, kwargs, return_value):
    strings = get_argument(args, kwargs, 0, "sequence")
    register_call("xml.etree.ElementTree.fromstringlist", "deserialize_op")

    for text in strings:
        extract_data_from_xml_body(user_input=text, root_element=return_value)


@on_import("xml.etree.ElementTree")
def patch(m):
    """
    patching module xml.etree.ElementTree
    - patches function fromstring(text, ...)
    - patches function fromstringlist(sequence, ...)
    (src: https://github.com/python/cpython/blob/bc1a6ecfab02075acea79f8460a2dce70c61b2fd/Lib/xml/etree/ElementTree.py#L1370)
    """
    patch_function(m, "fromstring", _fromstring)
    patch_function(m, "fromstringlist", _fromstringlist)



import pytest
from .is_private_ip import is_private_ip


# Test cases for is_private_ip
def test_private_ipv4_addresses():
    assert is_private_ip("192.168.1.1") is True
    assert is_private_ip("10.0.0.1") is True
    assert is_private_ip("172.16.0.1") is True
    assert is_private_ip("127.0.0.1") is True
    assert is_private_ip("169.254.1.1") is True


def test_public_ipv4_addresses():
    assert is_private_ip("8.8.8.8") is False
    assert is_private_ip("172.15.0.1") is False


def test_private_ipv6_addresses():
    assert is_private_ip("::1") is True  # Loopback address
    assert is_private_ip("fc00::1") is True  # Unique local address
    assert is_private_ip("fe80::1") is True  # Link-local address
    assert is_private_ip("2001:db8::1") is True  # Documentation address


def test_public_ipv6_addresses():
    assert is_private_ip("2a00::db8::1") is False  # Documentation address
    assert is_private_ip("::ffff:8.8.8.8") is False  # IPv4-mapped address


def test_invalid_addresses():
    assert is_private_ip("invalid-ip") is False
    assert is_private_ip("") is False
    assert is_private_ip("256.256.256.256") is False  # Invalid IPv4
    assert is_private_ip("::g") is False  # Invalid IPv6


def test_ipv4_mapped_ipv6():
    assert is_private_ip("::ffff:127.0.0.0") is True
    assert is_private_ip("::ffff:127.0.0.1") is True
    assert is_private_ip("::ffff:10.0.2.3") is True
    assert is_private_ip("::ffff:192.168.2.1") is True

    assert is_private_ip("::ffff:1.2.3.4") is False
    assert is_private_ip("::ffff:192.2.3.4") is False
    assert is_private_ip("::ffff:172.1.2.3") is False



"""Mainly exports on_detected_attack"""

import json

from aikido_zen.helpers.get_current_unixtime_ms import get_unixtime_ms
from aikido_zen.helpers.logging import logger
from aikido_zen.helpers.limit_length_metadata import limit_length_metadata
from aikido_zen.helpers.serialize_to_json import serialize_to_json


def on_detected_attack(connection_manager, attack, context, blocked, stack):
    """
    This will send something to the API when an attack is detected
    """
    if not connection_manager.token:
        return
    # Modify attack so we can send it out :
    try:
        attack["user"] = getattr(context, "user", None)
        attack["payload"] = json.dumps(attack["payload"])[:4096]
        attack["metadata"] = limit_length_metadata(attack["metadata"], 4096)
        attack["blocked"] = blocked
        attack["stack"] = stack

        payload = {
            "type": "detected_attack",
            "time": get_unixtime_ms(),
            "agent": connection_manager.get_manager_info(),
            "attack": attack,
            "request": {
                "method": context.method,
                "url": context.url,
                "ipAddress": context.remote_address,
                "userAgent": context.get_user_agent(),
                "body": context.body,
                "headers": context.headers,
                "source": context.source,
                "route": context.route,
            },
        }
        logger.debug(serialize_to_json(payload))
        result = connection_manager.api.report(
            connection_manager.token,
            payload,
            connection_manager.timeout_in_sec,
        )
        logger.debug("Result : %s", result)
    except Exception as e:
        logger.debug(e)
        logger.info("Failed to report an attack")



"""Exports ratelimiting and user blocking middleware for Flask"""

from aikido_zen.helpers.logging import logger
from .should_block_request import should_block_request


class AikidoFlaskMiddleware:
    """Ratelimiting and user blocking middleware for Flask"""

    def __init__(self, app):
        self.app = app
        try:
            from werkzeug.wrappers import Response

            self.Response = Response
        except ImportError:
            logger.warning(
                "Something went wrong whilst importing werkzeug.wrappers, middleware does not work"
            )

    def __call__(self, environ, start_response):
        result = should_block_request()
        if result["block"] is not True or self.Response is None:
            return self.app(environ, start_response)

        if result["type"] == "ratelimited":
            message = "You are rate limited by Zen."
            if result["trigger"] == "ip" and result["ip"]:
                message += " (Your IP: " + result["ip"] + ")"
            res = self.Response(message, mimetype="text/plain", status=429)
            return res(environ, start_response)
        elif result["type"] == "blocked":
            res = self.Response(
                "You are blocked by Zen.", mimetype="text/plain", status=403
            )
            return res(environ, start_response)
        logger.debug("Unknown type for blocking request: %s", result["type"])
        return self.app(environ, start_response)



"""
Mainly exports the LRUCache class
"""

from collections import OrderedDict
from aikido_zen.helpers.get_current_unixtime_ms import get_unixtime_ms


class LRUCache:
    """Least Recently Used cache"""

    def __init__(self, max_items, time_to_live_in_ms):
        if max_items <= 0:
            raise ValueError("Invalid max value")
        if time_to_live_in_ms < 0:
            raise ValueError("Invalid ttl value")

        self.max_items = max_items
        self.time_to_live_in_ms = time_to_live_in_ms
        self.cache = OrderedDict()

    def get(self, key):
        """Get a value from the cache"""
        if key in self.cache:
            # Check if the item is still valid based on TTL
            if (
                get_unixtime_ms(monotonic=True) - self.cache[key]["startTime"]
                < self.time_to_live_in_ms
            ):
                return self.cache[key]["value"]  # Return the actual value
            else:
                # Remove expired item
                del self.cache[key]
        return None

    def set(self, key, value):
        """Set a value in the cache based on the key"""
        if key in self.cache:
            del self.cache[key]  # Remove the existing item
        elif len(self.cache) >= self.max_items:
            self.cache.popitem(last=False)  # Remove the oldest item
        self.cache[key] = {
            "value": value,
            "startTime": get_unixtime_ms(monotonic=True),
        }  # Store value and timestamp

    def clear(self):
        """Clear the cache entirely"""
        self.cache.clear()

    def delete(self, key):
        """Delete a cache item using the key as id"""
        if key in self.cache:
            del self.cache[key]

    def keys(self):
        """Get the keys in the cache"""
        return list(self.cache.keys())

    @property
    def size(self):
        """Get the size of the cache"""
        return len(self.cache)



import pytest
from .detect_shell_injection import detect_shell_injection


def is_shell_injection(command: str, user_input: str):
    assert (
        detect_shell_injection(command, user_input) == True
    ), f"command: {command}, userInput: {user_input}"


def is_not_shell_injection(command: str, user_input: str):
    assert (
        detect_shell_injection(command, user_input) == False
    ), f"command: {command}, userInput: {user_input}"


def test_single_characters_ignored():
    is_not_shell_injection("ls `", "`")
    is_not_shell_injection("ls *", "*")
    is_not_shell_injection("ls a", "a")


def test_no_user_input():
    is_not_shell_injection("ls", "")
    is_not_shell_injection("ls", " ")
    is_not_shell_injection("ls", "  ")
    is_not_shell_injection("ls", "   ")


def test_user_input_not_in_command():
    is_not_shell_injection("ls", "$(echo)")


def test_user_input_longer_than_command():
    is_not_shell_injection("`ls`", "`ls` `ls`")


def test_detects_command_substitution():
    is_shell_injection("ls $(echo)", "$(echo)")
    is_shell_injection('ls "$(echo)"', "$(echo)")
    is_shell_injection(
        'echo $(echo "Inner: $(echo "This is nested")")',
        '$(echo "Inner: $(echo "This is nested")")',
    )

    is_not_shell_injection("ls '$(echo)'", "$(echo)")
    is_not_shell_injection(
        'ls \'$(echo "Inner: $(echo "This is nested")")\'',
        '$(echo "Inner: $(echo "This is nested")")',
    )


def test_detects_backticks():
    is_shell_injection("echo `echo`", "`echo`")


def test_checks_unsafely_quoted():
    is_shell_injection("ls '$(echo)", "$(echo)")


def test_single_quote_between_single_quotes():
    is_shell_injection("ls ''single quote''", "'single quote'")


# Test not ported : Requires `quote` function
# t.test("it ignores escaped backticks", async () => {
#  const domain = "www.example`whoami`.com";
#  const args = ["--domain", domain];
#  isNotShellInjection(`binary ${quote(args)}`, domain);
# });


def test_no_special_chars_inside_double_quotes():
    is_shell_injection('ls "whatever$"', "whatever$")
    is_shell_injection('ls "whatever!"', "whatever!")
    is_shell_injection('ls "whatever`"', "whatever`")


def test_no_semi_colon():
    is_shell_injection("ls whatever;", "whatever;")
    is_not_shell_injection('ls "whatever;"', "whatever;")
    is_not_shell_injection("ls 'whatever;'", "whatever;")


def test_rm_rf_executed_by_using_semi_colon():
    is_shell_injection("ls; rm -rf", "; rm -rf")


def test_rm_rf_is_flagged_as_shell_injection():
    is_shell_injection("rm -rf", "rm -rf")


def test_shell_injection_with_chained_commands_using_and():
    is_shell_injection("ls && rm -rf /", "&& rm -rf /")


def test_shell_injection_with_or_logic():
    is_shell_injection("ls || echo 'malicious code'", "|| echo 'malicious code'")


def test_detects_redirection_attempts():
    is_shell_injection("ls > /dev/null", "> /dev/null")
    is_shell_injection("cat file.txt > /etc/passwd", "> /etc/passwd")


def test_detects_append_redirection_attempts():
    is_shell_injection("echo 'data' >> /etc/passwd", ">> /etc/passwd")


def test_detects_pipe_character_as_shell_injection():
    is_shell_injection("cat file.txt | grep 'password'", "| grep 'password'")


def test_allows_safe_use_of_pipe_character_within_quotes():
    is_not_shell_injection("echo '|'", "|")


def test_detects_nested_command_substitution():
    is_shell_injection("echo $(cat $(ls))", "$(cat $(ls))")


def test_allows_safe_commands_within_single_quotes():
    is_not_shell_injection("echo 'safe command'", "safe command")


def test_detects_unsafe_use_of_variables():
    is_shell_injection("echo $USER", "$USER")
    is_shell_injection("echo ${USER}", "${USER}")
    is_shell_injection('echo "${USER}"', "${USER}")


def test_allows_safe_use_of_variables_within_quotes():
    is_not_shell_injection("echo '$USER'", "$USER")


def test_detects_subshell_execution_with_backticks_inside_double_quotes():
    is_shell_injection('ls "$(echo `whoami`)"', "`whoami`")


def test_detects_code_injection_with_newline_characters():
    is_shell_injection("echo 'safe'\necho 'malicious'", "\necho 'malicious'")


def test_detects_attempts_to_escape_out_of_quotes():
    is_shell_injection('echo "safe"; echo "malicious"', '"; echo "malicious"')


def test_correctly_handles_whitespace_in_inputs():
    is_not_shell_injection("ls", "   ")
    is_shell_injection("ls ; rm -rf /", "; rm -rf /")


def test_detects_file_manipulation_commands():
    is_shell_injection("touch /tmp/malicious", "touch /tmp/malicious")
    is_shell_injection("mv /tmp/safe /tmp/malicious", "mv /tmp/safe /tmp/malicious")


def test_allows_commands_with_constants_that_resemble_user_input():
    is_not_shell_injection("echo 'userInput'", "userInput")


def test_recognizes_safe_paths_that_include_patterns_similar_to_user_input():
    is_not_shell_injection(
        "ls /constant/path/without/user/input/", "/constant/path/without/user/input/"
    )


def test_acknowledges_safe_use_of_special_characters_when_properly_encapsulated():
    is_not_shell_injection('echo ";"', ";")
    is_not_shell_injection('echo "&&"', "&&")
    is_not_shell_injection('echo "||"', "||")


def test_treats_encapsulated_redirection_and_pipe_symbols_as_safe():
    is_not_shell_injection("echo 'data > file.txt'", "data > file.txt")
    is_not_shell_injection("echo 'find | grep'", "find | grep")


def test_recognizes_safe_inclusion_of_special_patterns_within_quotes_as_non_injections():
    is_not_shell_injection("echo '$(command)'", "$(command)")


def test_considers_constants_with_semicolons_as_safe_when_non_executable():
    is_not_shell_injection("echo 'text; more text'", "text; more text")


def test_acknowledges_commands_that_look_dangerous_but_are_safe_due_to_quoting():
    is_not_shell_injection("echo '; rm -rf /'", "; rm -rf /")
    is_not_shell_injection("echo '&& echo malicious'", "&& echo malicious")


def test_recognizes_commands_with_newline_characters_as_safe_when_encapsulated():
    is_not_shell_injection("echo 'line1\nline2'", "line1\nline2")


def test_accepts_special_characters_in_constants_as_safe_when_no_execution():
    is_not_shell_injection("echo '*'", "*")
    is_not_shell_injection("echo '?'", "?")
    is_not_shell_injection("echo '\\' ", "\\")


def test_does_not_flag_command_with_matching_whitespace_as_injection():
    is_not_shell_injection(
        "ls -l", " "
    )  # A single space is just an argument separator, not user input


def test_ignores_commands_where_multiple_spaces_match_user_input():
    is_not_shell_injection(
        "ls   -l", "   "
    )  # Multiple spaces between arguments should not be considered injection


def test_does_not_consider_leading_whitespace_in_commands_as_user_input():
    is_not_shell_injection(
        "  ls -l", "  "
    )  # Leading spaces before the command are not user-controlled


def test_treats_trailing_whitespace_in_commands_as_non_injection():
    is_not_shell_injection("ls -l ", " ")  # Trailing space after the command is benign


def test_recognizes_spaces_between_quotes_as_non_injective():
    is_not_shell_injection(
        "echo ' ' ", " "
    )  # Space within quotes is part of the argument, not a separator


def test_handles_spaces_within_quoted_arguments_correctly():
    is_not_shell_injection(
        "command 'arg with spaces'", " "
    )  # Spaces within a quoted argument should not be flagged


def test_correctly_interprets_spaces_in_mixed_argument_types():
    is_not_shell_injection(
        "command arg1 'arg with spaces' arg2", " "
    )  # Mixed argument types with internal spaces are safe


def test_ignores_spaces_in_commands_with_concatenated_arguments():
    is_not_shell_injection(
        "command 'arg1'arg2'arg3'", " "
    )  # Lack of spaces in concatenated arguments is intentional and safe


def test_does_not_flag_spaces_in_commands_with_no_arguments():
    is_not_shell_injection("command", " ")  # No arguments mean spaces are irrelevant


def test_considers_spaces_in_environment_variable_assignments_as_safe():
    is_not_shell_injection(
        "ENV_VAR='value' command", " "
    )  # Spaces around environment variable assignments are not injections


def test_new_lines_in_commands_are_considered_injections():
    is_shell_injection("ls \nrm", "\nrm")
    is_shell_injection("ls \nrm -rf", "\nrm -rf")


def test_new_lines_alone_are_not_considered_injections():
    is_not_shell_injection("ls\n\n", "\n\n")


def test_certain_commands_are_always_flagged_as_dangerous():
    is_shell_injection("/bin/rm -rf", "/bin/rm -rf")
    is_shell_injection("rm -rf", "rm -rf")
    is_shell_injection("rm -rf /", "rm -rf /")
    is_shell_injection("sleep 10", "sleep 10")
    is_shell_injection("sleep 10 &", "sleep 10 &")
    is_shell_injection("shutdown -h now", "shutdown -h now")
    is_shell_injection("halt", "halt")
    is_shell_injection("poweroff", "poweroff")
    is_shell_injection("reboot", "reboot")
    is_shell_injection("reboot -f", "reboot -f")
    is_shell_injection("ifconfig", "ifconfig")
    is_shell_injection("ifconfig -a", "ifconfig -a")
    is_shell_injection("kill", "kill")
    is_shell_injection("killall", "killall")
    is_shell_injection("killall -9", "killall -9")
    is_shell_injection("chmod", "chmod")
    is_shell_injection("chmod 777", "chmod 777")
    is_shell_injection("chown", "chown")
    is_shell_injection("chown root", "chown root")


def test_rm_being_part_of_other_commands():
    is_shell_injection('find /path/to/search -type f -name "pattern" | xargs rm', "rm")
    is_shell_injection(
        'find /path/to/search -type f -name "pattern" -exec rm {} \\;', "rm"
    )
    is_shell_injection("ls .|rm", "rm")


def test_ignores_dangerous_commands_if_part_of_string():
    is_not_shell_injection("binary sleepwithme", "sleepwithme")
    is_not_shell_injection("binary rm-rf", "rm-rf")
    is_not_shell_injection("term", "term")
    is_not_shell_injection("rm /files/rm.txt", "rm.txt")


def test_does_not_flag_domain_name_as_argument_unless_contains_backticks():
    is_not_shell_injection("binary --domain www.example.com", "www.example.com")
    is_not_shell_injection(
        "binary --domain https://www.example.com", "https://www.example.com"
    )

    is_shell_injection(
        "binary --domain www.example`whoami`.com", "www.example`whoami`.com"
    )
    is_shell_injection(
        "binary --domain https://www.example`whoami`.com",
        "https://www.example`whoami`.com",
    )


def test_flags_colon_if_used_as_command():
    is_shell_injection(":|echo", ":|")
    is_shell_injection(":| echo", ":|")
    is_shell_injection(": | echo", ": |")


def test_detects_shell_injection():
    is_shell_injection("/usr/bin/kill", "/usr/bin/kill")


def test_detects_shell_injection_with_uppercase_path():
    is_shell_injection("/usr/bIn/kill", "/usr/bIn/kill")


def test_detects_shell_injection_with_uppercase_command():
    is_shell_injection("/bin/CAT", "/bin/CAT")


def test_detects_shell_injection_with_uppercase_path_and_command():
    is_shell_injection("/bIn/LS -la", "/bIn/LS -la")


def test_shell_injection_with_multiple_slashes():
    is_shell_injection("//bin/ls", "//bin/ls")
    is_shell_injection("///bin/ls", "///bin/ls")


def test_shell_injection_with_dotdot():
    is_shell_injection("../bin/ls", "../bin/ls")
    is_shell_injection("../../bin/ls", "../../bin/ls")
    is_shell_injection("/../bin/ls", "/../bin/ls")
    is_shell_injection("/./bin/ls", "/./bin/ls")


def test_shell_injection_with_tilde():
    is_shell_injection("echo ~", "~")
    is_shell_injection("ls ~/.ssh", "~/.ssh")


def test_no_shell_injection_with_tilde():
    is_not_shell_injection("~", "~")
    is_not_shell_injection("ls ~/path", "path")


def test_real_case():
    is_shell_injection(
        "command -disable-update-check -target https://examplx.com|curl+https://cde-123.abc.domain.com+%23 -json-export /tmp/5891/8526757.json -tags microsoft,windows,exchange,iis,gitlab,oracle,cisco,joomla -stats -stats-interval 3 -retries 3 -no-stdin",
        "https://examplx.com|curl+https://cde-123.abc.domain.com+%23",
    )


def test_false_positive_with_email():
    is_not_shell_injection(
        "echo token | docker login --username john.doe@acme.com --password-stdin hub.acme.com",
        "john.doe@acme.com",
    )


def test_at_sign_with_shell_syntax():
    is_shell_injection("'echo \"${array[@]}\"'", "${array[@]}")
    is_shell_injection("echo $@", "$@")


def test_allows_comma_separated_list():
    is_not_shell_injection(
        "command -tags php,laravel,drupal,phpmyadmin,symfony -stats",
        "php,laravel,drupal,phpmyadmin,symfony",
    )


def test_it_flags_comma_in_loop():
    is_shell_injection(
        """command for (( i=0, j=10; i<j; i++, j-- ))
    do
        echo "$i $j"
    done""",
        "for (( i=0, j=10; i<j; i++, j-- ))",
    )



import pytest
from .check_context_for_shell_injection import check_context_for_shell_injection
from aikido_zen.context import Context


class Context1(Context):
    def __init__(self):
        self.cookies = {}
        self.headers = {}
        self.remote_address = "1.1.1.1"
        self.method = "POST"
        self.url = "url"
        self.query = {}
        self.body = {
            "domain": "www.example`whoami`.com",
        }
        self.source = "express"
        self.route = "/"
        self.parsed_userinput = {}


class Context2(Context):
    def __init__(self):
        self.cookies = {}
        self.headers = {}
        self.remote_address = "ip"
        self.method = "POST"
        self.url = "url"
        self.body = {}
        self.query = {
            "domain": "www.example`whoami`.com",
        }
        self.source = "express"
        self.route = "/"
        self.parsed_userinput = {}


def test_detect_shell_injection(monkeypatch):
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)

    context = Context1()
    result = check_context_for_shell_injection(
        command="binary --domain www.example`whoami`.com",
        operation="child_process.exec",
        context=context,
    )

    expected = {
        "operation": "child_process.exec",
        "kind": "shell_injection",
        "source": "body",
        "pathToPayload": ".domain",
        "metadata": {
            "command": "binary --domain www.example`whoami`.com",
        },
        "payload": "www.example`whoami`.com",
    }

    assert result == expected


def test_detect_shell_injection_from_route_params(monkeypatch):
    monkeypatch.setattr("aikido_zen.context.get_current_context", lambda: None)

    context = Context2()
    result = check_context_for_shell_injection(
        command="binary --domain www.example`whoami`.com",
        operation="child_process.exec",
        context=context,
    )

    expected = {
        "operation": "child_process.exec",
        "kind": "shell_injection",
        "source": "query",
        "pathToPayload": ".domain",
        "metadata": {
            "command": "binary --domain www.example`whoami`.com",
        },
        "payload": "www.example`whoami`.com",
    }

    assert result == expected


@pytest.mark.parametrize(
    "invalid_input",
    [
        None,
        123,  # Integer
        45.67,  # Float
        [],  # Empty list
        [1, 2, 3],  # List of integers
        {},  # Empty dictionary
        {"key": "value"},  # Dictionary
        set(),  # Empty set
        {1, 2, 3},  # Set of integers
        object(),  # Instance of a generic object
        lambda x: x,  # Lambda function
        (1, 2),  # Tuple
        b"bytes",  # Bytes
    ],
)
def test_doesnt_crash_with_invalid_command(invalid_input):
    context = Context2()
    result = check_context_for_shell_injection(
        command=invalid_input,
        operation="child_process.exec",
        context=context,
    )
    assert result == {}



"""
Mainly exports function `find_hostname_in_context`
"""

from aikido_zen.helpers.extract_strings_from_context import extract_strings_from_context
from .find_hostname_in_userinput import find_hostname_in_userinput
from .is_request_to_itself import is_request_to_itself
from ...context import Context


def find_hostname_in_context(hostname, context: Context, port):
    """Tries to locate the given hostname from context"""
    if not isinstance(hostname, str) or not isinstance(port, int):
        # Validate hostname and port input
        return None

    # We don't want to block outgoing requests to the same host as the server
    # (often happens that we have a match on headers like `Host`, `Origin`, `Referer`, etc.)
    if is_request_to_itself(context.url, hostname, port):
        return None

    # Punycode detected in hostname, while user input may not be in Punycode
    # We need to convert it to ensure we compare the right values
    if "xn--" in hostname:
        try:
            hostname = hostname.encode("ascii").decode("idna")
        except Exception:
            # Seems to be a malformed Punycode sequence, retain original
            # hostname
            pass

    for user_input, path, source in extract_strings_from_context(context):
        found = find_hostname_in_userinput(user_input, hostname, port)
        if found:
            return {
                "source": source,
                "pathToPayload": path,
                "payload": user_input,
            }



"""Exports function should_block_request"""

from aikido_zen.helpers.logging import logger
from aikido_zen.context import get_current_context
from aikido_zen.thread.thread_cache import get_cache
import aikido_zen.background_process.comms as c
from aikido_zen.ratelimiting.get_ratelimited_endpoint import get_ratelimited_endpoint
from aikido_zen.helpers.match_endpoints import match_endpoints


def should_block_request():
    """
    Checks for rate-limiting and checks if the current user is blocked.
    """
    try:
        context = get_current_context()
        cache = get_cache()
        if not context or not cache:
            return {"block": False}

        context.executed_middleware = (
            True  # Update context with middleware execution set to true
        )
        context.set_as_current_context()

        # Make sure we set middleware installed to true (reports back to core) :
        cache.middleware_installed = True

        # Blocked users:
        if context.user and cache.is_user_blocked(context.user["id"]):
            return {"block": True, "type": "blocked", "trigger": "user"}

        route_metadata = context.get_route_metadata()
        endpoints = cache.get_endpoints()
        comms = c.get_comms()
        if not comms or not endpoints:
            return {"block": False}
        matched_endpoints = match_endpoints(route_metadata, endpoints)
        # Ratelimiting :
        if matched_endpoints and get_ratelimited_endpoint(
            matched_endpoints, context.route
        ):
            # As an optimization check if the route is rate limited before sending over IPC
            ratelimit_res = comms.send_data_to_bg_process(
                action="SHOULD_RATELIMIT",
                obj={
                    "route_metadata": route_metadata,
                    "user": context.user,
                    "remote_address": context.remote_address,
                },
                receive=True,
            )
            if ratelimit_res["success"] and ratelimit_res["data"]["block"]:
                return {
                    "block": True,
                    "type": "ratelimited",
                    "trigger": ratelimit_res["data"]["trigger"],
                    "ip": context.remote_address,
                }
    except Exception as e:
        logger.debug("Exception occurred in should_block_request: %s", e)
    return {"block": False}



import pytest
from queue import Queue
from unittest.mock import MagicMock
from .attack import process_attack


class MockCloudConnectionManager:
    def __init__(self):
        self.statistics = MagicMock()


def test_process_attack_adds_data_to_queue():
    queue = Queue()
    connection_manager = MockCloudConnectionManager()
    data = ("injection_results", "context", True, "stacktrace")  # Example data
    process_attack(connection_manager, data, queue)

    # Check if the data is added to the queue
    assert not queue.empty()
    assert queue.get() == data


def test_process_attack_statistics_not_called_when_disabled():
    queue = Queue()
    connection_manager = MockCloudConnectionManager()
    connection_manager.statistics = None  # Disable statistics
    data = ("injection_results", "context", True, "stacktrace")  # Example data
    process_attack(connection_manager, data, queue)

    # Check if on_detected_attack was not called
    assert (
        connection_manager.statistics is None
        or not connection_manager.statistics.on_detected_attack.called
    )


def test_process_attack_multiple_calls():
    queue = Queue()
    connection_manager = MockCloudConnectionManager()
    data1 = ("injection_results_1", "context_1", True, "stacktrace_1")
    data2 = ("injection_results_2", "context_2", False, "stacktrace_2")

    process_attack(connection_manager, data1, queue)
    process_attack(connection_manager, data2, queue)

    # Check if both data items are added to the queue
    assert queue.qsize() == 2
    assert queue.get() == data1
    assert queue.get() == data2


def test_process_attack_with_different_data_formats():
    queue = Queue()
    connection_manager = MockCloudConnectionManager()

    # Test with different types of data
    data1 = ("injection_results", "context", True, "stacktrace")
    data2 = ("injection_results", "context", False, "stacktrace")
    data3 = ("injection_results", "context", None, "stacktrace")

    process_attack(connection_manager, data1, queue)
    process_attack(connection_manager, data2, queue)
    process_attack(connection_manager, data3, queue)

    # Check if all data items are added to the queue
    assert queue.qsize() == 3
    assert queue.get() == data1
    assert queue.get() == data2
    assert queue.get() == data3



""" 
init.py file for api/ folder. Includes abstract class ReportingApi
"""

import json
from aikido_zen.helpers.logging import logger


class ReportingApi:
    """This is the super class for the reporting API's"""

    def to_api_response(self, res):
        """Converts results into an Api response obj"""
        status = res.status_code
        if status == 429:
            return {"success": False, "error": "rate_limited"}
        elif status == 401:
            return {"success": False, "error": "invalid_token"}
        elif status == 200:
            try:
                return json.loads(res.text)
            except Exception as e:
                logger.debug(e)
                logger.debug(res.text)
        return {"success": False, "error": "unknown_error"}

    def report(self, token, event, timeout_in_sec):
        """Report event to aikido server"""

    def fetch_firewall_lists(self, token):
        """Fetches the different lists from aikido's servers"""



import pytest
from .contains_shell_syntax import contains_shell_syntax


def test_detects_shell_syntax():
    assert contains_shell_syntax("", "") is False
    assert contains_shell_syntax("hello", "hello") is False
    assert contains_shell_syntax("\n", "\n") is False
    assert contains_shell_syntax("\n\n", "\n\n") is False

    assert contains_shell_syntax("$(command)", "$(command)") is True
    assert contains_shell_syntax("$(command arg arg)", "$(command arg arg)") is True
    assert contains_shell_syntax("`command`", "`command`") is True
    assert contains_shell_syntax("\narg", "\narg") is True
    assert contains_shell_syntax("\targ", "\targ") is True
    assert contains_shell_syntax("\narg\n", "\narg\n") is True
    assert contains_shell_syntax("arg\n", "arg\n") is True
    assert contains_shell_syntax("arg\narg", "arg\narg") is True
    assert contains_shell_syntax("rm -rf", "rm -rf") is True
    assert contains_shell_syntax("/bin/rm -rf", "/bin/rm -rf") is True
    assert contains_shell_syntax("/bin/rm", "/bin/rm") is True
    assert contains_shell_syntax("/sbin/sleep", "/sbin/sleep") is True
    assert contains_shell_syntax("/usr/bin/kill", "/usr/bin/kill") is True
    assert contains_shell_syntax("/usr/bin/killall", "/usr/bin/killall") is True
    assert contains_shell_syntax("/usr/bin/env", "/usr/bin/env") is True
    assert contains_shell_syntax("/bin/ps", "/bin/ps") is True
    assert contains_shell_syntax("/usr/bin/W", "/usr/bin/W") is True
    assert contains_shell_syntax("lsattr", "lsattr") is True


def test_detects_commands_surrounded_by_separators():
    assert (
        contains_shell_syntax(
            r'find /path/to/search -type f -name "pattern" -exec rm {} \\;', "rm"
        )
        is True
    )


def test_detects_commands_with_separator_before():
    assert (
        contains_shell_syntax(
            'find /path/to/search -type f -name "pattern" | xargs rm', "rm"
        )
        is True
    )


def test_detects_commands_with_separator_after():
    assert contains_shell_syntax("rm arg", "rm") is True


def test_checks_if_same_command_occurs_in_user_input():
    assert contains_shell_syntax("find cp", "rm") is False


def test_treats_colon_as_command():
    assert contains_shell_syntax(":|echo", ":|") is True
    assert (
        contains_shell_syntax("https://www.google.com", "https://www.google.com")
        is False
    )


def test_detects_commands_with_separators():
    assert contains_shell_syntax("rm>arg", "rm") is True


def test_detects_commands_with_separators():
    assert contains_shell_syntax("rm<arg", "rm") is True


def test_empty_command_and_input():
    assert contains_shell_syntax("", "") is False
    assert contains_shell_syntax("", "rm") is False
    assert contains_shell_syntax("rm", "") is False


def test_command_with_special_characters():
    assert contains_shell_syntax("echo $HOME", "echo") is True
    assert contains_shell_syntax("echo $HOME", "$HOME") is True
    assert contains_shell_syntax('echo "Hello World"', "echo") is True
    assert contains_shell_syntax("echo 'Hello World'", "echo") is True


def test_command_with_multiple_separators():
    assert contains_shell_syntax("rm -rf; echo 'done'", "rm") is True
    assert contains_shell_syntax("ls | grep 'test'", "ls") is True
    assert contains_shell_syntax("find . -name '*.txt' | xargs rm", "rm") is True


def test_command_with_path_prefixes():
    assert contains_shell_syntax("/bin/rm -rf /tmp", "/bin/rm") is True
    assert (
        contains_shell_syntax("/usr/bin/killall process_name", "/usr/bin/killall")
        is True
    )
    assert contains_shell_syntax("/sbin/shutdown now", "/sbin/shutdown") is True


def test_command_with_colon():
    assert contains_shell_syntax(":; echo 'test'", ":") is True
    assert contains_shell_syntax("echo :; echo 'test'", ":") is True


def test_command_with_newline_separators():
    assert contains_shell_syntax("echo 'Hello'\nrm -rf /tmp", "rm") is True
    assert contains_shell_syntax("echo 'Hello'\n", "echo") is True


def test_command_with_tabs():
    assert contains_shell_syntax("echo 'Hello'\trm -rf /tmp", "rm") is True
    assert contains_shell_syntax("\techo 'Hello'", "echo") is True


def test_command_with_invalid_input():
    assert contains_shell_syntax("echo 'Hello'", "invalid_command") is False
    assert contains_shell_syntax("ls -l", "rm") is False


def test_command_with_multiple_commands():
    assert contains_shell_syntax("rm -rf; ls -l; echo 'done'", "ls") is True
    assert contains_shell_syntax("echo 'Hello'; rm -rf /tmp", "rm") is True


def test_command_with_no_separators():
    assert contains_shell_syntax("echoHello", "echo") is False
    assert contains_shell_syntax("rmrf", "rm") is False


def test_command_with_dangerous_chars():
    assert contains_shell_syntax("rm -rf; echo 'done'", ";") is True
    assert contains_shell_syntax("echo 'Hello' & rm -rf /tmp", "&") is True
    assert contains_shell_syntax("echo 'Hello' | rm -rf /tmp", "|") is True


def test_command_with_path_and_arguments():
    assert contains_shell_syntax("/usr/bin/ls -l", "/usr/bin/ls") is True
    assert contains_shell_syntax("/bin/cp file1 file2", "/bin/cp") is True



"""
The code to send out a heartbeat is in here
"""

from aikido_zen.helpers.logging import logger
from aikido_zen.helpers.create_interval import create_interval


def send_heartbeats_every_x_secs(connection_manager, interval_in_secs, event_scheduler):
    """
    Start sending out heartbeats every x seconds
    """
    if connection_manager.serverless:
        logger.debug("Running in serverless environment, not starting heartbeats")
        return
    if not connection_manager.token:
        logger.debug("No token provided, not starting heartbeats")
        return

    logger.debug("Starting heartbeats")

    # Create an interval for "interval_in_secs" seconds :
    create_interval(
        event_scheduler=event_scheduler,
        interval_in_secs=interval_in_secs,
        function=lambda connection_manager: connection_manager.send_heartbeat(),
        args=(connection_manager,),
    )



"""Exports the function starts_with_unsafe_path"""

linux_root_folders = [
    "/bin/",
    "/boot/",
    "/dev/",
    "/etc/",
    "/home/",
    "/init/",
    "/lib/",
    "/media/",
    "/mnt/",
    "/opt/",
    "/proc/",
    "/root/",
    "/run/",
    "/sbin/",
    "/srv/",
    "/sys/",
    "/tmp/",
    "/usr/",
    "/var/",
]

# List of dangerous path starts, including Windows paths
dangerous_path_starts = linux_root_folders + ["c:/", "c:\\"]


def starts_with_unsafe_path(file_path, user_input):
    """Check if the file path starts with any dangerous paths and the user input."""
    lower_case_path = file_path.lower()
    lower_case_user_input = user_input.lower()

    for dangerous_start in dangerous_path_starts:
        if lower_case_path.startswith(dangerous_start) and lower_case_path.startswith(
            lower_case_user_input
        ):
            return True

    return False



"""
Helper function file, see function docstring
"""


def limit_length_metadata(metadata, max_length):
    """
    Limits the length of the metadata obj so it can be sent out
    """
    for key in metadata:
        if len(metadata[key]) > max_length:
            metadata[key] = metadata[key][:max_length]

    return metadata



class Users:
    def __init__(self, max_entries=1000):
        self.max_entries = max_entries
        self.users = {}

    def add_user(self, user_id, user_name, user_ip, current_time):
        self.ensure_max_entries()

        first_seen_at = current_time
        if self.users.get(user_id):
            # Use the first_seen_at timestamp of the existing user
            first_seen_at = self.users.get(user_id).get("firstSeenAt")

        self.users[user_id] = {
            "id": user_id,
            "name": user_name,
            "lastIpAddress": user_ip,
            "firstSeenAt": first_seen_at,
            "lastSeenAt": current_time,
        }

    def add_user_from_entry(self, user_entry):
        self.ensure_max_entries()

        existing_user = self.users.get(user_entry["id"])
        if existing_user:
            # Use the firstSeenAt timestamp of the existing user
            user_entry["firstSeenAt"] = existing_user["firstSeenAt"]

        self.users[user_entry["id"]] = user_entry

    def ensure_max_entries(self):
        if len(self.users) >= self.max_entries:
            # Remove the first added user (FIFO)
            first_added_key = next(iter(self.users))
            del self.users[first_added_key]

    def as_array(self):
        return [dict(user) for user in self.users.values()]

    def clear(self):
        self.users.clear()



import pytest
from .detect_path_traversal import detect_path_traversal


def test_empty_user_input():
    assert detect_path_traversal("test.txt", "") is False


def test_empty_file_input():
    assert detect_path_traversal("", "test") is False


def test_empty_user_input_and_file_input():
    assert detect_path_traversal("", "") is False


def test_user_input_is_a_single_character():
    assert detect_path_traversal("test.txt", "t") is False


def test_file_input_is_a_single_character():
    assert detect_path_traversal("t", "test") is False


def test_same_as_user_input():
    assert detect_path_traversal("text.txt", "text.txt") is False


def test_with_directory_before():
    assert detect_path_traversal("directory/text.txt", "text.txt") is False


def test_with_both_directory_before():
    assert detect_path_traversal("directory/text.txt", "directory/text.txt") is False


def test_user_input_and_file_input_are_single_characters():
    assert detect_path_traversal("t", "t") is False


def test_it_flags_dot_dot_slash():
    assert detect_path_traversal("../test.txt", "../") is True


def test_it_flags_backslash_dot_dot():
    assert detect_path_traversal("..\\test.txt", "..\\") is True


def test_it_flags_double_dot_slash():
    assert detect_path_traversal("../../test.txt", "../../") is True


def test_it_flags_double_dot_backslash():
    assert detect_path_traversal("..\\..\\test.txt", "..\\..\\") is True


def test_it_flags_four_dot_slash():
    assert detect_path_traversal("../../../../test.txt", "../../../../") is True


def test_it_flags_three_dot_backslash():
    assert detect_path_traversal("..\\..\\..\\test.txt", "..\\..\\..\\") is True


def test_user_input_is_longer_than_file_path():
    assert detect_path_traversal("../file.txt", "../../file.txt") is False


def test_absolute_linux_path():
    assert detect_path_traversal("/etc/passwd", "/etc/passwd") is True


def test_linux_user_directory():
    assert detect_path_traversal("/home/user/file.txt", "/home/user/") is True


def test_windows_drive_letter():
    assert detect_path_traversal("C:\\file.txt", "C:\\") is True


def test_no_path_traversal():
    assert (
        detect_path_traversal("/appdata/storage/file.txt", "/storage/file.txt") is False
    )


def test_does_not_flag_test():
    assert detect_path_traversal("/app/test.txt", "test") is False


def test_does_not_flag_example_test_txt():
    assert (
        detect_path_traversal("/app/data/example/test.txt", "example/test.txt") is False
    )


def test_does_not_absolute_path_with_different_folder():
    assert detect_path_traversal("/etc/app/config", "/etc/hack/config") is False


def test_does_not_absolute_path_inside_another_folder():
    assert detect_path_traversal("/etc/app/data/etc/config", "/etc/config") is False


def test_disable_checkPathStart():
    assert detect_path_traversal("/etc/passwd", "/etc/passwd", False) is False
